"""
@author: linlin

@references: 

	[1] Liva Ralaivola, Sanjay J Swamidass, Hiroto Saigo, and Pierre 
	Baldi. Graph kernels for chemical informatics. Neural networks, 
	18(8):1093–1110, 2005.
"""

import sys
import time
from collections import Counter
from itertools import chain
from functools import partial
from multiprocessing import Pool
from tqdm import tqdm

import networkx as nx
import numpy as np

from gklearn.utils.graphdataset import get_dataset_attributes
from gklearn.utils.parallel import parallel_gm
from gklearn.utils.trie import Trie


def untilhpathkernel(*args,
					 node_label='atom',
					 edge_label='bond_type',
					 depth=10,
					 k_func='MinMax',
					 compute_method='trie',
					 parallel='imap_unordered',
					 n_jobs=None,
					 chunksize=None,
					 verbose=True):
	"""Compute path graph kernels up to depth/hight h between graphs.
	
	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.
	
	G1, G2 : NetworkX graphs
		Two graphs between which the kernel is computed.

	node_label : string
		Node attribute used as label. The default node label is atom.

	edge_label : string
		Edge attribute used as label. The default edge label is bond_type.

	depth : integer
		Depth of search. Longest length of paths.

	k_func : function
		A kernel function applied using different notions of fingerprint 
		similarity, defining the type of feature map and normalization method 
		applied for the graph kernel. The Following choices are available:

		'MinMax': use the MiniMax kernel and counting feature map.

		'tanimoto': use the Tanimoto kernel and binary feature map.

		None: no sub-kernel is used, the kernel is computed directly.

	compute_method : string
		Computation method to store paths and compute the graph kernel. The 
		Following choices are available:

		'trie': store paths as tries.

		'naive': store paths to lists.

	n_jobs : int
		Number of jobs for parallelization.

	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the path kernel up to h between
		2 praphs.
	"""
	# pre-process
	depth = int(depth)
	Gn = args[0] if len(args) == 1 else [args[0], args[1]]
	Gn = [g.copy() for g in Gn]
	Kmatrix = np.zeros((len(Gn), len(Gn)))
	ds_attrs = get_dataset_attributes(
		Gn,
		attr_names=['node_labeled', 'node_attr_dim', 'edge_labeled', 
					'edge_attr_dim', 'is_directed'],
		node_label=node_label, edge_label=edge_label)
	if k_func is not None:
		if not ds_attrs['node_labeled']:
			for G in Gn:
				nx.set_node_attributes(G, '0', 'atom')
		if not ds_attrs['edge_labeled']:
			for G in Gn:
				nx.set_edge_attributes(G, '0', 'bond_type')

	start_time = time.time()		

	if parallel == 'imap_unordered':
		# ---- use pool.imap_unordered to parallel and track progress. ----
		# get all paths of all graphs before computing kernels to save time,
		# but this may cost a lot of memory for large datasets.
		pool = Pool(n_jobs)
		itr = zip(Gn, range(0, len(Gn)))
		if chunksize is None:
			if len(Gn) < 100 * n_jobs:
				chunksize = int(len(Gn) / n_jobs) + 1
			else:
				chunksize = 100
		all_paths = [[] for _ in range(len(Gn))]
		if compute_method == 'trie' and k_func is not None:
			getps_partial = partial(wrapper_find_all_path_as_trie, depth, 
									ds_attrs, node_label, edge_label)
		elif compute_method != 'trie' and k_func is not None:  
			getps_partial = partial(wrapper_find_all_paths_until_length, depth, 
									ds_attrs, node_label, edge_label, True)  
		else: 
			getps_partial = partial(wrapper_find_all_paths_until_length, depth, 
									ds_attrs, node_label, edge_label, False)
		if verbose:
			iterator = tqdm(pool.imap_unordered(getps_partial, itr, chunksize),
							desc='getting paths', file=sys.stdout)
		else:
			iterator = pool.imap_unordered(getps_partial, itr, chunksize)
		for i, ps in iterator:
			all_paths[i] = ps
		pool.close()
		pool.join()
	
#	for g in Gn:
#		if compute_method == 'trie' and k_func is not None:
#			find_all_path_as_trie(g, depth, ds_attrs, node_label, edge_label)
#		elif compute_method != 'trie' and k_func is not None:  
#			find_all_paths_until_length(g, depth, ds_attrs, node_label, edge_label)
#		else: 
#			find_all_paths_until_length(g, depth, ds_attrs, node_label, edge_label, False)
		
##	size = sys.getsizeof(all_paths)
##	for item in all_paths:
##		size += sys.getsizeof(item)
##		for pppps in item:
##			size += sys.getsizeof(pppps)
##	print(size)
#			
##	ttt = time.time()
##	# ---- ---- use pool.map to parallel ----
##	for i, ps in tqdm(
##			pool.map(getps_partial, range(0, len(Gn))),
##			desc='getting paths', file=sys.stdout):
##		all_paths[i] = ps
##	print(time.time() - ttt)
	 
		if compute_method == 'trie' and k_func is not None:
			def init_worker(trie_toshare):
				global G_trie
				G_trie = trie_toshare
			do_partial = partial(wrapper_uhpath_do_trie, k_func)
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
						glbv=(all_paths,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose) 
		elif compute_method != 'trie' and k_func is not None:
			def init_worker(plist_toshare):
				global G_plist
				G_plist = plist_toshare
			do_partial = partial(wrapper_uhpath_do_naive, k_func)   
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
						glbv=(all_paths,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose) 
		else:
			def init_worker(plist_toshare):
				global G_plist
				G_plist = plist_toshare
			do_partial = partial(wrapper_uhpath_do_kernelless, ds_attrs, edge_kernels)   
			parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
						glbv=(all_paths,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose) 
	
	elif parallel is None:
#		from pympler import asizeof
		# ---- direct running, normally use single CPU core. ----
#		print(asizeof.asized(all_paths, detail=1).format())
	
		if compute_method == 'trie':
			all_paths = [
				find_all_path_as_trie(Gn[i],
					 depth,
					 ds_attrs,
					 node_label=node_label,
					 edge_label=edge_label) for i in tqdm(
						range(0, len(Gn)), desc='getting paths', file=sys.stdout)
			]
#			sizeof_allpaths = asizeof.asizeof(all_paths)
#			print(sizeof_allpaths)
			pbar = tqdm(
				total=((len(Gn) + 1) * len(Gn) / 2),
				desc='Computing kernels',
				file=sys.stdout)
			for i in range(0, len(Gn)):
				for j in range(i, len(Gn)):
					Kmatrix[i][j] = _untilhpathkernel_do_trie(all_paths[i], 
						   all_paths[j], k_func)
					Kmatrix[j][i] = Kmatrix[i][j]
					pbar.update(1)
		else:
			all_paths = [
				find_all_paths_until_length(
					Gn[i],
					depth,
					ds_attrs,
					node_label=node_label,
					edge_label=edge_label) for i in tqdm(
						range(0, len(Gn)), desc='getting paths', file=sys.stdout)
			]
#			sizeof_allpaths = asizeof.asizeof(all_paths)
#			print(sizeof_allpaths)
			pbar = tqdm(
				total=((len(Gn) + 1) * len(Gn) / 2),
				desc='Computing kernels',
				file=sys.stdout)
			for i in range(0, len(Gn)):
				for j in range(i, len(Gn)):
					Kmatrix[i][j] = _untilhpathkernel_do_naive(all_paths[i], all_paths[j],
														 k_func)
					Kmatrix[j][i] = Kmatrix[i][j]
					pbar.update(1)

	run_time = time.time() - start_time
	if verbose:
		print("\n --- kernel matrix of path kernel up to %d of size %d built in %s seconds ---"
			  % (depth, len(Gn), run_time))

#	print(Kmatrix[0][0:10])
	return Kmatrix, run_time


def _untilhpathkernel_do_trie(trie1, trie2, k_func):
	"""Compute path graph kernels up to depth d between 2 graphs using trie.

	Parameters
	----------
	trie1, trie2 : list
		Tries that contains all paths in 2 graphs.
	k_func : function
		A kernel function applied using different notions of fingerprint 
		similarity.

	Return
	------
	kernel : float
		Path kernel up to h between 2 graphs.
	"""
	if k_func == 'tanimoto':	  
		# traverse all paths in graph1 and search them in graph2. Deep-first 
		# search is applied.
		def traverseTrie1t(root, trie2, setlist, pcurrent=[]):
			for key, node in root['children'].items():
				pcurrent.append(key)
				if node['isEndOfWord']:					
					setlist[1] += 1
					count2 = trie2.searchWord(pcurrent)
					if count2 != 0:
						setlist[0] += 1
				if node['children'] != {}:
					traverseTrie1t(node, trie2, setlist, pcurrent)
				else:
					del pcurrent[-1]
			if pcurrent != []:
				del pcurrent[-1]
				
				
		# traverse all paths in graph2 and find out those that are not in 
		# graph1. Deep-first search is applied. 
		def traverseTrie2t(root, trie1, setlist, pcurrent=[]):
			for key, node in root['children'].items():
				pcurrent.append(key)
				if node['isEndOfWord']:
		#					print(node['count'])
					count1 = trie1.searchWord(pcurrent)
					if count1 == 0:	
						setlist[1] += 1
				if node['children'] != {}:
					traverseTrie2t(node, trie1, setlist, pcurrent)
				else:
					del pcurrent[-1]
			if pcurrent != []:
				del pcurrent[-1]
		
		setlist = [0, 0] # intersection and union of path sets of g1, g2.
#		print(trie1.root)
#		print(trie2.root)
		traverseTrie1t(trie1.root, trie2, setlist)
#		print(setlist)
		traverseTrie2t(trie2.root, trie1, setlist)
#		print(setlist)
		kernel = setlist[0] / setlist[1]
		
	else: # MinMax kernel		  
		# traverse all paths in graph1 and search them in graph2. Deep-first 
		# search is applied.
		def traverseTrie1m(root, trie2, sumlist, pcurrent=[]):
			for key, node in root['children'].items():
				pcurrent.append(key)
				if node['isEndOfWord']:
		#					print(node['count'])
					count1 = node['count']
					count2 = trie2.searchWord(pcurrent)
					sumlist[0] += min(count1, count2)
					sumlist[1] += max(count1, count2)
				if node['children'] != {}:
					traverseTrie1m(node, trie2, sumlist, pcurrent)
				else:
					del pcurrent[-1]
			if pcurrent != []:
				del pcurrent[-1]
		
		# traverse all paths in graph2 and find out those that are not in 
		# graph1. Deep-first search is applied.				
		def traverseTrie2m(root, trie1, sumlist, pcurrent=[]):
			for key, node in root['children'].items():
				pcurrent.append(key)
				if node['isEndOfWord']:				   
		#					print(node['count'])
					count1 = trie1.searchWord(pcurrent)
					if count1 == 0:	
						sumlist[1] += node['count']
				if node['children'] != {}:
					traverseTrie2m(node, trie1, sumlist, pcurrent)
				else:
					del pcurrent[-1]
			if pcurrent != []:
				del pcurrent[-1]
		
		sumlist = [0, 0] # sum of mins and sum of maxs
#		print(trie1.root)
#		print(trie2.root)
		traverseTrie1m(trie1.root, trie2, sumlist)
#		print(sumlist)
		traverseTrie2m(trie2.root, trie1, sumlist)
#		print(sumlist)
		kernel = sumlist[0] / sumlist[1]

	return kernel


def wrapper_uhpath_do_trie(k_func, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _untilhpathkernel_do_trie(G_trie[i], G_trie[j], k_func)
		

def _untilhpathkernel_do_naive(paths1, paths2, k_func):
	"""Compute path graph kernels up to depth d between 2 graphs naively.

	Parameters
	----------
	paths_list : list of list
		List of list of paths in all graphs, where for unlabeled graphs, each 
		path is represented by a list of nodes; while for labeled graphs, each 
		path is represented by a string consists of labels of nodes and/or 
		edges on that path.
	k_func : function
		A kernel function applied using different notions of fingerprint 
		similarity.

	Return
	------
	kernel : float
		Path kernel up to h between 2 graphs.
	"""
	all_paths = list(set(paths1 + paths2))

	if k_func == 'tanimoto':
		length_union = len(set(paths1 + paths2))
		kernel = (len(set(paths1)) + len(set(paths2)) -
				  length_union) / length_union
#		vector1 = [(1 if path in paths1 else 0) for path in all_paths]
#		vector2 = [(1 if path in paths2 else 0) for path in all_paths]
#		kernel_uv = np.dot(vector1, vector2)
#		kernel = kernel_uv / (len(set(paths1)) + len(set(paths2)) - kernel_uv)

	else:  # MinMax kernel
		path_count1 = Counter(paths1)
		path_count2 = Counter(paths2)
		vector1 = [(path_count1[key] if (key in path_count1.keys()) else 0)
				   for key in all_paths]
		vector2 = [(path_count2[key] if (key in path_count2.keys()) else 0)
				   for key in all_paths]
		kernel = np.sum(np.minimum(vector1, vector2)) / \
			np.sum(np.maximum(vector1, vector2))

	return kernel


def wrapper_uhpath_do_naive(k_func, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _untilhpathkernel_do_naive(G_plist[i], G_plist[j], k_func)


def _untilhpathkernel_do_kernelless(paths1, paths2, k_func):
	"""Compute path graph kernels up to depth d between 2 graphs naively.

	Parameters
	----------
	paths_list : list of list
		List of list of paths in all graphs, where for unlabeled graphs, each 
		path is represented by a list of nodes; while for labeled graphs, each 
		path is represented by a string consists of labels of nodes and/or 
		edges on that path.
	k_func : function
		A kernel function applied using different notions of fingerprint 
		similarity.

	Return
	------
	kernel : float
		Path kernel up to h between 2 graphs.
	"""
	all_paths = list(set(paths1 + paths2))

	if k_func == 'tanimoto':
		length_union = len(set(paths1 + paths2))
		kernel = (len(set(paths1)) + len(set(paths2)) -
				  length_union) / length_union
#		vector1 = [(1 if path in paths1 else 0) for path in all_paths]
#		vector2 = [(1 if path in paths2 else 0) for path in all_paths]
#		kernel_uv = np.dot(vector1, vector2)
#		kernel = kernel_uv / (len(set(paths1)) + len(set(paths2)) - kernel_uv)

	else:  # MinMax kernel
		path_count1 = Counter(paths1)
		path_count2 = Counter(paths2)
		vector1 = [(path_count1[key] if (key in path_count1.keys()) else 0)
				   for key in all_paths]
		vector2 = [(path_count2[key] if (key in path_count2.keys()) else 0)
				   for key in all_paths]
		kernel = np.sum(np.minimum(vector1, vector2)) / \
			np.sum(np.maximum(vector1, vector2))

	return kernel


def wrapper_uhpath_do_kernelless(k_func, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _untilhpathkernel_do_kernelless(G_plist[i], G_plist[j], k_func)


# @todo: (can be removed maybe)  this method find paths repetively, it could be faster.
def find_all_paths_until_length(G,
								length,
								ds_attrs,
								node_label='atom',
								edge_label='bond_type',
								tolabelseqs=True):
	"""Find all paths no longer than a certain maximum length in a graph. A 
	recursive depth first search is applied.

	Parameters
	----------
	G : NetworkX graphs
		The graph in which paths are searched.
	length : integer
		The maximum length of paths.
	ds_attrs: dict
		Dataset attributes.
	node_label : string
		Node attribute used as label. The default node label is atom.
	edge_label : string
		Edge attribute used as label. The default edge label is bond_type.

	Return
	------
	path : list
		List of paths retrieved, where for unlabeled graphs, each path is 
		represented by a list of nodes; while for labeled graphs, each path is 
		represented by a list of strings consists of labels of nodes and/or 
		edges on that path.
	"""
	# path_l = [tuple([n]) for n in G.nodes]  # paths of length l
	# all_paths = path_l[:]
	# for l in range(1, length + 1):
	#	 path_l_new = []
	#	 for path in path_l:
	#		 for neighbor in G[path[-1]]:
	#			 if len(path) < 2 or neighbor != path[-2]:
	#				 tmp = path + (neighbor, )
	#				 if tuple(tmp[::-1]) not in path_l_new:
	#					 path_l_new.append(tuple(tmp))

	#	 all_paths += path_l_new
	#	 path_l = path_l_new[:]

	path_l = [[n] for n in G.nodes]  # paths of length l
	all_paths = [p.copy() for p in path_l]
	for l in range(1, length + 1):
		path_lplus1 = []
		for path in path_l:
			for neighbor in G[path[-1]]:
				if neighbor not in path:
					tmp = path + [neighbor]
#					if tmp[::-1] not in path_lplus1:
					path_lplus1.append(tmp)

		all_paths += path_lplus1
		path_l = [p.copy() for p in path_lplus1]

	# for i in range(0, length + 1):
	#	 new_paths = find_all_paths(G, i)
	#	 if new_paths == []:
	#		 break
	#	 all_paths.extend(new_paths)

	# consider labels
#	print(paths2labelseqs(all_paths, G, ds_attrs, node_label, edge_label))
#	print()
	return (paths2labelseqs(all_paths, G, ds_attrs, node_label, edge_label) 
			if tolabelseqs else all_paths)
		
		
def wrapper_find_all_paths_until_length(length, ds_attrs, node_label, 
									 edge_label, tolabelseqs, itr_item):
	g = itr_item[0]
	i = itr_item[1]
	return i, find_all_paths_until_length(g, length, ds_attrs,
				node_label=node_label, edge_label=edge_label, 
				tolabelseqs=tolabelseqs)


def find_all_path_as_trie(G,
						 length,
						 ds_attrs,
						 node_label='atom',
						 edge_label='bond_type'):
#	time1 = time.time()
	
#	all_path = find_all_paths_until_length(G, length, ds_attrs, 
#										   node_label=node_label,
#										   edge_label=edge_label)
#	ptrie = Trie()
#	for path in all_path:
#		ptrie.insertWord(path)
	
#	ptrie = Trie()
#	path_l = [[n] for n in G.nodes]  # paths of length l
#	path_l_str = paths2labelseqs(path_l, G, ds_attrs, node_label, edge_label)
#	for p in path_l_str:
#		ptrie.insertWord(p)
#	for l in range(1, length + 1):
#		path_lplus1 = []
#		for path in path_l:
#			for neighbor in G[path[-1]]:
#				if neighbor not in path:
#					tmp = path + [neighbor]
##					if tmp[::-1] not in path_lplus1:
#					path_lplus1.append(tmp)
#		path_l = path_lplus1[:]
#		# consider labels
#		path_l_str = paths2labelseqs(path_l, G, ds_attrs, node_label, edge_label)
#		for p in path_l_str:
#			ptrie.insertWord(p)
#	
#	print(time.time() - time1)
#	print(ptrie.root)
#	print()
			
			
	# traverse all paths up to length h in a graph and construct a trie with 
	# them. Deep-first search is applied. Notice the reverse of each path is 
	# also stored to the trie.			   
	def traverseGraph(root, ptrie, length, G, ds_attrs, node_label, edge_label,
					  pcurrent=[]):
		if len(pcurrent) < length + 1:
			for neighbor in G[root]:
				if neighbor not in pcurrent:
					pcurrent.append(neighbor)
					plstr = paths2labelseqs([pcurrent], G, ds_attrs, 
											node_label, edge_label)
					ptrie.insertWord(plstr[0])
					traverseGraph(neighbor, ptrie, length, G, ds_attrs, 
								   node_label, edge_label, pcurrent)
		del pcurrent[-1]


	ptrie = Trie()
	path_l = [[n] for n in G.nodes]  # paths of length l
	path_l_str = paths2labelseqs(path_l, G, ds_attrs, node_label, edge_label)
	for p in path_l_str:
		ptrie.insertWord(p)
	for n in G.nodes:
		traverseGraph(n, ptrie, length, G, ds_attrs, node_label, edge_label, 
					   pcurrent=[n])
		
		
#	def traverseGraph(root, all_paths, length, G, ds_attrs, node_label, edge_label,
#					  pcurrent=[]):
#		if len(pcurrent) < length + 1:
#			for neighbor in G[root]:
#				if neighbor not in pcurrent:
#					pcurrent.append(neighbor)
#					plstr = paths2labelseqs([pcurrent], G, ds_attrs, 
#											node_label, edge_label)
#					all_paths.append(pcurrent[:])
#					traverseGraph(neighbor, all_paths, length, G, ds_attrs, 
#								   node_label, edge_label, pcurrent)
#		del pcurrent[-1]
#
#
#	path_l = [[n] for n in G.nodes]  # paths of length l
#	all_paths = path_l[:]
#	path_l_str = paths2labelseqs(path_l, G, ds_attrs, node_label, edge_label)
##	for p in path_l_str:
##		ptrie.insertWord(p)
#	for n in G.nodes:
#		traverseGraph(n, all_paths, length, G, ds_attrs, node_label, edge_label, 
#					   pcurrent=[n])
	
#	print(ptrie.root)
	return ptrie


def wrapper_find_all_path_as_trie(length, ds_attrs, node_label, 
									 edge_label, itr_item):
	g = itr_item[0]
	i = itr_item[1]
	return i, find_all_path_as_trie(g, length, ds_attrs,
				node_label=node_label, edge_label=edge_label)


def paths2labelseqs(plist, G, ds_attrs, node_label, edge_label):
	if ds_attrs['node_labeled']:
		if ds_attrs['edge_labeled']:
			path_strs = [
				tuple(
					list(
						chain.from_iterable(
							(G.nodes[node][node_label],
							 G[node][path[idx + 1]][edge_label])
							for idx, node in enumerate(path[:-1]))) +
					[G.nodes[path[-1]][node_label]]) for path in plist
			]
			# path_strs = []
			# for path in all_paths:
			#	 strlist = list(
			#		 chain.from_iterable((G.node[node][node_label],
			#							  G[node][path[idx + 1]][edge_label])
			#							 for idx, node in enumerate(path[:-1])))
			#	 strlist.append(G.node[path[-1]][node_label])
			#	 path_strs.append(tuple(strlist))
		else:
			path_strs = [
				tuple([G.nodes[node][node_label] for node in path])
				for path in plist
			]
		return path_strs
	else:
		if ds_attrs['edge_labeled']:
			return [
				tuple([] if len(path) == 1 else [
					G[node][path[idx + 1]][edge_label]
					for idx, node in enumerate(path[:-1])
				]) for path in plist
			]
		else:
			return [tuple(['0' for node in path]) for path in plist]
#			return [tuple([len(path)]) for path in all_paths]   

#
#def paths2GSuffixTree(paths):
#	return Tree(paths, builder=ukkonen.Builder)


# def find_paths(G, source_node, length):
#	 """Find all paths no longer than a certain length those start from a source node. A recursive depth first search is applied.

#	 Parameters
#	 ----------
#	 G : NetworkX graphs
#		 The graph in which paths are searched.
#	 source_node : integer
#		 The number of the node from where all paths start.
#	 length : integer
#		 The length of paths.

#	 Return
#	 ------
#	 path : list of list
#		 List of paths retrieved, where each path is represented by a list of nodes.
#	 """
#	 return [[source_node]] if length == 0 else \
#		 [[source_node] + path for neighbor in G[source_node]
#		  for path in find_paths(G, neighbor, length - 1) if source_node not in path]

# def find_all_paths(G, length):
#	 """Find all paths with a certain length in a graph. A recursive depth first search is applied.

#	 Parameters
#	 ----------
#	 G : NetworkX graphs
#		 The graph in which paths are searched.
#	 length : integer
#		 The length of paths.

#	 Return
#	 ------
#	 path : list of list
#		 List of paths retrieved, where each path is represented by a list of nodes.
#	 """
#	 all_paths = []
#	 for node in G:
#		 all_paths.extend(find_paths(G, node, length))

#	 # The following process is not carried out according to the original article
#	 # all_paths_r = [ path[::-1] for path in all_paths ]

#	 # # For each path, two presentation are retrieved from its two extremities. Remove one of them.
#	 # for idx, path in enumerate(all_paths[:-1]):
#	 #	 for path2 in all_paths_r[idx+1::]:
#	 #		 if path == path2:
#	 #			 all_paths[idx] = []
#	 #			 break

#	 # return list(filter(lambda a: a != [], all_paths))
#	 return all_paths



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 23 09:52:50 2020

@author: ljia
"""
import time

class Timer(object):
	"""A timer class that can be used by methods that support time limits.
	
	Note
	----
	This is the Python implementation of `the C++ code in GEDLIB <https://github.com/dbblumenthal/gedlib/blob/master/src/env/timer.hpp>`__.
	"""
	
	def __init__(self, time_limit_in_sec):
		"""Constructs a timer for a given time limit.
		
		Parameters
		----------
		time_limit_in_sec : string
			The time limit in seconds.
		"""		
		self._time_limit_in_sec = time_limit_in_sec
		self._start_time = time.time()
	
	
	def expired(self):
		"""Checks if the time limit has expired. 
		
		Return
		------
		Boolean true if the time limit has expired and false otherwise.
"""
		if self._time_limit_in_sec > 0:
			runtime = time.time() - self._start_time
			return runtime >= self._time_limit_in_sec
		return False


"""
run_wlsubtree.py



@Author: linlin
@Date: 24.01.24
"""
import os


def main():
	import sys
	import pathlib
	sys.path.insert(0, "../../../")

	import numpy as np

	from gklearn.utils.model_selection_precomputed_new import \
		model_selection_for_precomputed_kernel_new
	from datasets.ds import dslist

	import multiprocessing

	from gklearn.kernels.weisfeiler_lehman import WLSubtree

	cur_dir = os.path.dirname(os.path.realpath(__file__)) + '/'

	dslist = [
		#    {'name': 'Alkane', 'dataset': '../datasets/Alkane/dataset.ds', 'task': 'regression',
		#        'dataset_y': '../datasets/Alkane/dataset_boiling_point_names.txt'},
		#    # contains single node graph, node symb
		#    {'name': 'Acyclic', 'dataset': '../datasets/acyclic/dataset_bps.ds',
		#        'task': 'regression'},  # node symb
		#    {'name': 'MAO', 'dataset': '../datasets/MAO/dataset.ds'}, # node/edge symb
		#    {'name': 'PAH', 'dataset': '../datasets/PAH/dataset.ds'}, # unlabeled
		{'name': 'MUTAG', 'dataset': cur_dir + '../../../../datasets/MUTAG/MUTAG_A.txt'},
		# node/edge symb
		#     {'name': 'Letter-med', 'dataset': '../datasets/Letter-med/Letter-med_A.txt'},
		# node nsymb
		#    {'name': 'AIDS', 'dataset': '../datasets/AIDS/AIDS_A.txt'}, # node symb/nsymb, edge symb
		#    {'name': 'ENZYMES', 'dataset': '../datasets/ENZYMES_txt/ENZYMES_A_sparse.txt'},
		# node symb/nsymb
		#    {'name': 'NCI1', 'dataset': '../datasets/NCI1/NCI1_A.txt'}, # node symb
		#    {'name': 'NCI109', 'dataset': '../datasets/NCI109/NCI109_A.txt'}, # node symb
		#    {'name': 'D&D', 'dataset': '../datasets/DD/DD_A.txt'}, # node symb
		#    {'name': 'monoterpenoides', 'dataset': '../datasets/monoterpenoides/dataset_10+.ds'}, # node/edge symb

		#
		#    {'name': 'Mutagenicity', 'dataset': '../datasets/Mutagenicity/Mutagenicity_A.txt'},
		#    # node/edge symb
		#     {'name': 'COIL-DEL', 'dataset': '../datasets/COIL-DEL/COIL-DEL_A.txt'}, # edge symb, node nsymb
		# # #     {'name': 'BZR', 'dataset': '../datasets/BZR_txt/BZR_A_sparse.txt'}, # node symb/nsymb
		# # #     {'name': 'COX2', 'dataset': '../datasets/COX2_txt/COX2_A_sparse.txt'}, # node symb/nsymb
		#     {'name': 'Fingerprint', 'dataset': '../datasets/Fingerprint/Fingerprint_A.txt'},
		#
		# #     {'name': 'DHFR', 'dataset': '../datasets/DHFR_txt/DHFR_A_sparse.txt'}, # node symb/nsymb
		# #     {'name': 'SYNTHETIC', 'dataset': '../datasets/SYNTHETIC_txt/SYNTHETIC_A_sparse.txt'}, # node symb/nsymb
		# #     {'name': 'MSRC9', 'dataset': '../datasets/MSRC_9_txt/MSRC_9_A.txt'}, # node symb
		# #     {'name': 'MSRC21', 'dataset': '../datasets/MSRC_21_txt/MSRC_21_A.txt'}, # node symb
		# #     {'name': 'FIRSTMM_DB', 'dataset': '../datasets/FIRSTMM_DB/FIRSTMM_DB_A.txt'}, # node symb/nsymb ,edge nsymb

		# #     {'name': 'PROTEINS', 'dataset': '../datasets/PROTEINS_txt/PROTEINS_A_sparse.txt'}, # node symb/nsymb
		# #     {'name': 'PROTEINS_full', 'dataset': '../datasets/PROTEINS_full_txt/PROTEINS_full_A_sparse.txt'}, # node symb/nsymb
		#     {'name': 'NCI-HIV', 'dataset': '../datasets/NCI-HIV/AIDO99SD.sdf',
		#         'dataset_y': '../datasets/NCI-HIV/aids_conc_may04.txt',}, # node/edge symb

		#     # not working below
		#     {'name': 'PTC_FM', 'dataset': '../datasets/PTC/Train/FM.ds',},
		#     {'name': 'PTC_FR', 'dataset': '../datasets/PTC/Train/FR.ds',},
		#     {'name': 'PTC_MM', 'dataset': '../datasets/PTC/Train/MM.ds',},
		#     {'name': 'PTC_MR', 'dataset': '../datasets/PTC/Train/MR.ds',},
	]
	estimator = WLSubtree
	param_grid_precomputed = {
		# 'base_kernel': ['subtree'],
		'height': np.linspace(0, 10, 11)
	}
	param_grid = [{'C': np.logspace(-10, 4, num=29, base=10)},
	              {'alpha': np.logspace(-10, 10, num=41, base=10)}]

	for ds in dslist:
		print()
		print(ds['name'])
		model_selection_for_precomputed_kernel_new(
			ds['dataset'],
			estimator,
			param_grid_precomputed,
			(param_grid[1] if ('task' in ds and ds['task']
			                   == 'regression') else param_grid[0]),
			(ds['task'] if 'task' in ds else 'classification'),
			NUM_TRIALS=30, # todo: 30
			datafile_y=(ds['dataset_y'] if 'dataset_y' in ds else None),
			extra_params=(ds['extra_params'] if 'extra_params' in ds else None),
			ds_name=ds['name'],
			n_jobs=1,# todo:multiprocessing.cpu_count(),
			read_gm_from_file=False,
			verbose=True
		)
		print()


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  6 12:08:24 2020

@author: ljia
"""
import random
import numpy as np

def test_get_nb_edit_operations_symbolic_cml():
	"""Test get_nb_edit_operations_symbolic_cml().
	"""
	"""**1.   Get dataset.**"""

	from gklearn.utils import Dataset
	
	# Predefined dataset name, use dataset "MUTAG".
	ds_name = 'MUTAG'
	
	# Initialize a Dataset.
	dataset = Dataset()
	# Load predefined dataset "MUTAG".
	dataset.load_predefined_dataset(ds_name)
	graph1 = dataset.graphs[0]
	graph2 = dataset.graphs[1]
	
	"""**2.  Compute graph edit distance.**"""
	
# 	try:
	# Initialize label costs randomly.
	node_label_costs, edge_label_costs = _initialize_label_costs(dataset)
	
	# Compute GEDs.
	pi_forward, pi_backward, dis, node_labels, edge_labels = _compute_ged(dataset, node_label_costs, edge_label_costs)
	
	
	# Compute numbers of edit operations.
	
	from gklearn.ged.util.util import get_nb_edit_operations_symbolic_cml
	
	n_edit_operations = get_nb_edit_operations_symbolic_cml(graph1, graph2, pi_forward, pi_backward, node_labels, edge_labels)
	
	assert np.abs((np.dot(np.concatenate((node_label_costs, edge_label_costs)), n_edit_operations) - dis) / dis) < 10e-6
	
# 	except Exception as exception:
# 		assert False, exception
		
		
def _initialize_label_costs(dataset):
	node_label_costs = _initialize_node_label_costs(dataset)
	edge_label_costs = _initialize_edge_label_costs(dataset)
	return node_label_costs, edge_label_costs
	
	
def _initialize_node_label_costs(dataset):
	# Get list of node labels.
	nls = dataset.get_all_node_labels()
	# Generate random costs.
	nb_nl = int((len(nls) * (len(nls) - 1)) / 2 + 2 * len(nls))
	rand_costs = random.sample(range(1, 10 * nb_nl + 1), nb_nl)
	rand_costs /= np.max(rand_costs)
			
	return rand_costs


def _initialize_edge_label_costs(dataset):
	# Get list of edge labels.
	els = dataset.get_all_edge_labels()
	# Generate random costs.
	nb_el = int((len(els) * (len(els) - 1)) / 2 + 2 * len(els))
	rand_costs = random.sample(range(1, 10 * nb_el + 1), nb_el)
	rand_costs /= np.max(rand_costs)
			
	return rand_costs


def _compute_ged(dataset, node_label_costs, edge_label_costs):
	from gklearn.ged.env import GEDEnv
	from gklearn.ged.util.util import label_costs_to_matrix
	import networkx as nx
			
	ged_env = GEDEnv() # initailize GED environment.
	ged_env.set_edit_cost('CONSTANT', # GED cost type.
	                      edit_cost_constants=[3, 3, 1, 3, 3, 1] # edit costs.
						  )
	for g in dataset.graphs:
		ged_env.add_nx_graph(g, '') # add graphs

	node_labels = ged_env.get_all_node_labels()
	edge_labels = ged_env.get_all_edge_labels()
	listID = ged_env.get_all_graph_ids() # get list IDs of graphs
	ged_env.set_label_costs(label_costs_to_matrix(node_label_costs, len(node_labels)), 
					  label_costs_to_matrix(edge_label_costs, len(edge_labels)))
	ged_env.init(init_type='LAZY_WITHOUT_SHUFFLED_COPIES') # initialize GED environment.
	options = {'initialization_method': 'RANDOM', # or 'NODE', etc.
	           'threads': 1 # parallel threads.
			   }
	ged_env.set_method('BIPARTITE', # GED method.
	                   options # options for GED method.
					   )
	ged_env.init_method() # initialize GED method.
	
	ged_env.run_method(listID[0], listID[1]) # run.
	
	pi_forward = ged_env.get_forward_map(listID[0], listID[1]) # forward map.
	pi_backward = ged_env.get_backward_map(listID[0], listID[1]) # backward map.
	dis = ged_env.get_upper_bound(listID[0], listID[1])	# GED bewteen two graphs.
	
	# make the map label correct (label remove map as np.inf)
	nodes1 = [n for n in dataset.graphs[0].nodes()]
	nodes2 = [n for n in dataset.graphs[1].nodes()]
	nb1 = nx.number_of_nodes(dataset.graphs[0])
	nb2 = nx.number_of_nodes(dataset.graphs[1])
	pi_forward = [nodes2[pi] if pi < nb2 else np.inf for pi in pi_forward]
	pi_backward = [nodes1[pi] if pi < nb1 else np.inf for pi in pi_backward]
	
	return pi_forward, pi_backward, dis, node_labels, edge_labels
		

if __name__ == "__main__":
	test_get_nb_edit_operations_symbolic_cml()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  3 22:22:57 2020

@author: ljia

@references:

	[1] H. Kashima, K. Tsuda, and A. Inokuchi. Marginalized kernels between
	labeled graphs. In Proceedings of the 20th International Conference on
	Machine Learning, Washington, DC, United States, 2003.

	[2] Pierre Mahé, Nobuhisa Ueda, Tatsuya Akutsu, Jean-Luc Perret, and
	Jean-Philippe Vert. Extensions of marginalized graph kernels. In
	Proceedings of the twenty-first international conference on Machine
	learning, page 70. ACM, 2004.
"""

import sys
from multiprocessing import Pool
from gklearn.utils import get_iters
import numpy as np
import networkx as nx
from gklearn.utils import SpecialLabel
from gklearn.utils.kernels import deltakernel
from gklearn.utils.parallel import parallel_gm, parallel_me
from gklearn.utils.utils import untotterTransformation
from gklearn.kernels import GraphKernel


class Marginalized(GraphKernel):

	def __init__(self, **kwargs):
		GraphKernel.__init__(self)
		self._node_labels = kwargs.get('node_labels', [])
		self._edge_labels = kwargs.get('edge_labels', [])
		self._p_quit = kwargs.get('p_quit', 0.5)
		self._n_iteration = kwargs.get('n_iteration', 10)
		self._remove_totters = kwargs.get('remove_totters', False)
		self._ds_infos = kwargs.get('ds_infos', {})
		self._n_iteration = int(self._n_iteration)


	def _compute_gm_series(self, graphs):
		self._add_dummy_labels(graphs)

		if self._remove_totters:
			iterator = get_iters(graphs, desc='removing tottering', file=sys.stdout, verbose=(self.verbose >= 2))
			# @todo: this may not work.
			graphs = [untotterTransformation(G, self._node_labels, self._edge_labels) for G in iterator]

		# compute Gram matrix.
		gram_matrix = np.zeros((len(graphs), len(graphs)))

		from itertools import combinations_with_replacement
		itr = combinations_with_replacement(range(0, len(graphs)), 2)
		len_itr = int(len(graphs) * (len(graphs) + 1) / 2)
		iterator = get_iters(itr, desc='Computing kernels', file=sys.stdout,
					length=len_itr, verbose=(self.verbose >= 2))
		for i, j in iterator:
			kernel = self._kernel_do(graphs[i], graphs[j])
			gram_matrix[i][j] = kernel
			gram_matrix[j][i] = kernel # @todo: no directed graph considered?

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		self._add_dummy_labels(self._graphs)

		if self._remove_totters:
			pool = Pool(self.n_jobs)
			itr = range(0, len(self._graphs))
			if len(self._graphs) < 100 * self.n_jobs:
				chunksize = int(len(self._graphs) / self.n_jobs) + 1
			else:
				chunksize = 100
			remove_fun = self._wrapper_untotter
			iterator = get_iters(pool.imap_unordered(remove_fun, itr, chunksize),
							desc='removing tottering', file=sys.stdout,
							length=len(self._graphs), verbose=(self.verbose >= 2))
			for i, g in iterator:
				self._graphs[i] = g
			pool.close()
			pool.join()

		# compute Gram matrix.
		gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))

		def init_worker(gn_toshare):
			global G_gn
			G_gn = gn_toshare
		do_fun = self._wrapper_kernel_do
		parallel_gm(do_fun, gram_matrix, self._graphs, init_worker=init_worker,
					glbv=(self._graphs,), n_jobs=self.n_jobs, verbose=self.verbose)

		return gram_matrix


	def _compute_kernel_list_series(self, g1, g_list):
		self._add_dummy_labels(g_list + [g1])

		if self._remove_totters:
			g1 = untotterTransformation(g1, self._node_labels, self._edge_labels) # @todo: this may not work.
			iterator = get_iters(g_list, desc='removing tottering', file=sys.stdout, verbose=(self.verbose >= 2))
			# @todo: this may not work.
			g_list = [untotterTransformation(G, self._node_labels, self._edge_labels) for G in iterator]

		# compute kernel list.
		kernel_list = [None] * len(g_list)
		iterator = get_iters(range(len(g_list)), desc='Computing kernels', file=sys.stdout, length=len(g_list), verbose=(self.verbose >= 2))
		for i in iterator:
			kernel = self._kernel_do(g1, g_list[i])
			kernel_list[i] = kernel

		return kernel_list


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._add_dummy_labels(g_list + [g1])

		if self._remove_totters:
			g1 = untotterTransformation(g1, self._node_labels, self._edge_labels) # @todo: this may not work.
			pool = Pool(self.n_jobs)
			itr = range(0, len(g_list))
			if len(g_list) < 100 * self.n_jobs:
				chunksize = int(len(g_list) / self.n_jobs) + 1
			else:
				chunksize = 100
			remove_fun = self._wrapper_untotter
			iterator = get_iters(pool.imap_unordered(remove_fun, itr, chunksize),
							desc='removing tottering', file=sys.stdout,
							length=len(g_list), verbose=(self.verbose >= 2))
			for i, g in iterator:
				g_list[i] = g
			pool.close()
			pool.join()

		# compute kernel list.
		kernel_list = [None] * len(g_list)

		def init_worker(g1_toshare, g_list_toshare):
			global G_g1, G_g_list
			G_g1 = g1_toshare
			G_g_list = g_list_toshare
		do_fun = self._wrapper_kernel_list_do
		def func_assign(result, var_to_assign):
			var_to_assign[result[0]] = result[1]
		itr = range(len(g_list))
		len_itr = len(g_list)
		parallel_me(do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
			init_worker=init_worker, glbv=(g1, g_list), method='imap_unordered',
			n_jobs=self.n_jobs, itr_desc='Computing kernels', verbose=self.verbose)

		return kernel_list


	def _wrapper_kernel_list_do(self, itr):
		return itr, self._kernel_do(G_g1, G_g_list[itr])


	def _compute_single_kernel_series(self, g1, g2):
		self._add_dummy_labels([g1] + [g2])
		if self._remove_totters:
			g1 = untotterTransformation(g1, self._node_labels, self._edge_labels) # @todo: this may not work.
			g2 = untotterTransformation(g2, self._node_labels, self._edge_labels)
		kernel = self._kernel_do(g1, g2)
		return kernel


	def _kernel_do(self, g1, g2):
		"""Compute marginalized graph kernel between 2 graphs.

		Parameters
		----------
		g1, g2 : NetworkX graphs
			2 graphs between which the kernel is computed.

		Return
		------
		kernel : float
			Marginalized kernel between 2 graphs.
		"""
		# init parameters
		kernel = 0
		num_nodes_G1 = nx.number_of_nodes(g1)
		num_nodes_G2 = nx.number_of_nodes(g2)
		# the initial probability distribution in the random walks generating step
		# (uniform distribution over |G|)
		p_init_G1 = 1 / num_nodes_G1
		p_init_G2 = 1 / num_nodes_G2

		q = self._p_quit * self._p_quit
		r1 = q

	#	# initial R_inf
	#	# matrix to save all the R_inf for all pairs of nodes
	#	R_inf = np.zeros([num_nodes_G1, num_nodes_G2])
	#
	#	# Compute R_inf with a simple interative method
	#	for i in range(1, n_iteration):
	#		R_inf_new = np.zeros([num_nodes_G1, num_nodes_G2])
	#		R_inf_new.fill(r1)
	#
	#		# Compute R_inf for each pair of nodes
	#		for node1 in g1.nodes(data=True):
	#			neighbor_n1 = g1[node1[0]]
	#			# the transition probability distribution in the random walks
	#			# generating step (uniform distribution over the vertices adjacent
	#			# to the current vertex)
	#			if len(neighbor_n1) > 0:
	#				p_trans_n1 = (1 - p_quit) / len(neighbor_n1)
	#				for node2 in g2.nodes(data=True):
	#					neighbor_n2 = g2[node2[0]]
	#					if len(neighbor_n2) > 0:
	#						p_trans_n2 = (1 - p_quit) / len(neighbor_n2)
	#
	#						for neighbor1 in neighbor_n1:
	#							for neighbor2 in neighbor_n2:
	#								t = p_trans_n1 * p_trans_n2 * \
	#									deltakernel(g1.node[neighbor1][node_label],
	#												g2.node[neighbor2][node_label]) * \
	#									deltakernel(
	#										neighbor_n1[neighbor1][edge_label],
	#										neighbor_n2[neighbor2][edge_label])
	#
	#								R_inf_new[node1[0]][node2[0]] += t * R_inf[neighbor1][
	#									neighbor2]  # ref [1] equation (8)
	#		R_inf[:] = R_inf_new
	#
	#	# add elements of R_inf up and compute kernel
	#	for node1 in g1.nodes(data=True):
	#		for node2 in g2.nodes(data=True):
	#			s = p_init_G1 * p_init_G2 * deltakernel(
	#				node1[1][node_label], node2[1][node_label])
	#			kernel += s * R_inf[node1[0]][node2[0]]  # ref [1] equation (6)


		R_inf = {} # dict to save all the R_inf for all pairs of nodes
		# initial R_inf, the 1st iteration.
		for node1 in g1.nodes():
			for node2 in g2.nodes():
	#			R_inf[(node1[0], node2[0])] = r1
				if len(g1[node1]) > 0:
					if len(g2[node2]) > 0:
						R_inf[(node1, node2)] = r1
					else:
						R_inf[(node1, node2)] = self._p_quit
				else:
					if len(g2[node2]) > 0:
						R_inf[(node1, node2)] = self._p_quit
					else:
						R_inf[(node1, node2)] = 1

		# compute all transition probability first.
		t_dict = {}
		if self._n_iteration > 1:
			for node1 in g1.nodes():
				neighbor_n1 = g1[node1]
				# the transition probability distribution in the random walks
				# generating step (uniform distribution over the vertices adjacent
				# to the current vertex)
				if len(neighbor_n1) > 0:
					p_trans_n1 = (1 - self._p_quit) / len(neighbor_n1)
					for node2 in g2.nodes():
						neighbor_n2 = g2[node2]
						if len(neighbor_n2) > 0:
							p_trans_n2 = (1 - self._p_quit) / len(neighbor_n2)
							for neighbor1 in neighbor_n1:
								for neighbor2 in neighbor_n2:
									t_dict[(node1, node2, neighbor1, neighbor2)] = \
										p_trans_n1 * p_trans_n2 * \
										deltakernel(tuple(g1.nodes[neighbor1][nl] for nl in self._node_labels), tuple(g2.nodes[neighbor2][nl] for nl in self._node_labels)) * \
										deltakernel(tuple(neighbor_n1[neighbor1][el] for el in self._edge_labels), tuple(neighbor_n2[neighbor2][el] for el in self._edge_labels))

		# Compute R_inf with a simple interative method
		for i in range(2, self._n_iteration + 1):
			R_inf_old = R_inf.copy()

			# Compute R_inf for each pair of nodes
			for node1 in g1.nodes():
				neighbor_n1 = g1[node1]
				# the transition probability distribution in the random walks
				# generating step (uniform distribution over the vertices adjacent
				# to the current vertex)
				if len(neighbor_n1) > 0:
					for node2 in g2.nodes():
						neighbor_n2 = g2[node2]
						if len(neighbor_n2) > 0:
							R_inf[(node1, node2)] = r1
							for neighbor1 in neighbor_n1:
								for neighbor2 in neighbor_n2:
									R_inf[(node1, node2)] += \
										(t_dict[(node1, node2, neighbor1, neighbor2)] * \
										R_inf_old[(neighbor1, neighbor2)])  # ref [1] equation (8)

		# add elements of R_inf up and compute kernel.
		for (n1, n2), value in R_inf.items():
			s = p_init_G1 * p_init_G2 * deltakernel(tuple(g1.nodes[n1][nl] for nl in self._node_labels), tuple(g2.nodes[n2][nl] for nl in self._node_labels))
			kernel += s * value  # ref [1] equation (6)

		return kernel


	def _wrapper_kernel_do(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._kernel_do(G_gn[i], G_gn[j])


	def _wrapper_untotter(self, i):
		return i, untotterTransformation(self._graphs[i], self._node_labels, self._edge_labels) # @todo: this may not work.


	def _add_dummy_labels(self, Gn):
		if len(self._node_labels) == 0 or (len(self._node_labels) == 1 and self._node_labels[0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_node_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self._node_labels = [SpecialLabel.DUMMY]
		if len(self._edge_labels) == 0 or (len(self._edge_labels) == 1 and self._edge_labels[0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_edge_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self._edge_labels = [SpecialLabel.DUMMY]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct  7 14:43:36 2020

@author: ljia
"""

def rounder(x, decimals):
	"""Round, where 5 is rounded up.

	Parameters
	----------
	x : float
		The number to be rounded.
	decimals : int
		Decimals to which ``x'' is rounded.

	Returns
	-------
	string
		The rounded number.
	"""
	x_strs = str(x).split('.')
	if len(x_strs) == 2:
		before = x_strs[0]
		after = x_strs[1]
		if len(after) > decimals:
			if int(after[decimals]) >= 5:
				after0s = ''
				for c in after:
					if c == '0':
						after0s += '0'
					elif c != '0':
						break
				if len(after0s) == decimals:
					after0s = after0s[:-1]
				after = after0s + str(int(after[0:decimals]) + 1)[-decimals:]
			else:
				after = after[0:decimals]
		elif len(after) < decimals:
			after += '0' * (decimals - len(after))
		return before + '.' + after

	elif len(x_strs) == 1:
		return x_strs[0]
	
	
if __name__ == '__main__':
	x = 1.0075333616
	y = rounder(x, 2)
	print(y)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 19 18:13:56 2020

@author: ljia
"""
from gklearn.utils import dummy_node


def construct_node_map_from_solver(solver, node_map, solution_id):
	node_map.clear()
	num_nodes_g = node_map.num_source_nodes()
	num_nodes_h = node_map.num_target_nodes()
	
	# add deletions and substitutions
	for row in range(0, num_nodes_g):
		col = solver.get_assigned_col(row, solution_id)
		if col >= num_nodes_h:
			node_map.add_assignment(row, dummy_node())
		else:
			node_map.add_assignment(row, col)
			
	# insertions.
	for col in range(0, num_nodes_h):
		if solver.get_assigned_row(col, solution_id) >= num_nodes_g:
			node_map.add_assignment(dummy_node(), col)
	

def options_string_to_options_map(options_string):
    """Transforms an options string into an options map.
    
    Parameters
    ----------
    options_string : string
        Options string of the form "[--<option> <arg>] [...]".
    
    Return
    ------
    options_map : dict{string : string}
        Map with one key-value pair (<option>, <arg>) for each option contained in the string.
    """
    if options_string == '':
        return
    options_map = {}
    words = []
    tokenize(options_string, ' ', words)
    expect_option_name = True
    for word in words:
        if expect_option_name:
            is_opt_name, word = is_option_name(word)
            if is_opt_name:
                option_name = word
                if option_name in options_map:
                    raise Exception('Multiple specification of option "' + option_name + '".')
                options_map[option_name] = ''
            else:
                raise Exception('Invalid options "' + options_string + '". Usage: options = "[--<option> <arg>] [...]"')
        else:
            is_opt_name, word = is_option_name(word)
            if is_opt_name:
                raise Exception('Invalid options "' + options_string + '". Usage: options = "[--<option> <arg>] [...]"')
            else:
                options_map[option_name] = word
        expect_option_name = not expect_option_name
    return options_map
    

def tokenize(sentence, sep, words):
    """Separates a sentence into words separated by sep (unless contained in single quotes).
    
    Parameters
    ----------
    sentence : string
        The sentence that should be tokenized.
        
    sep : string 
        The separator. Must be different from "'".
        
    words : list[string]
        The obtained words.
    """
    outside_quotes = True
    word_length = 0
    pos_word_start = 0
    for pos in range(0, len(sentence)):
        if sentence[pos] == '\'':
            if not outside_quotes and pos < len(sentence) - 1:
                if sentence[pos + 1] != sep:
                    raise Exception('Sentence contains closing single quote which is followed by a char different from ' + sep + '.')
            word_length += 1
            outside_quotes = not outside_quotes
        elif outside_quotes and sentence[pos] == sep:
            if word_length > 0:
                words.append(sentence[pos_word_start:pos_word_start + word_length])
            pos_word_start = pos + 1
            word_length = 0
        else:
            word_length += 1
    if not outside_quotes:
        raise Exception('Sentence contains unbalanced single quotes.')
    if word_length > 0:
        words.append(sentence[pos_word_start:pos_word_start + word_length])


def is_option_name(word):
    """Checks whether a word is an option name and, if so, removes the leading dashes.
    
    Parameters
    ----------
    word : string
        Word.
        
    return
    ------
    True if word is of the form "--<option>".
    
    word : string
        The word without the leading dashes.
    """
    if word[0] == '\'':
        word = word[1:len(word) - 2]
        return False, word
    if len(word) < 3:
        return False, word
    if word[0] == '-' and word[1] == '-' and word[2] != '-':
        word = word[2:]
        return True, word
    return False, word


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  5 09:42:30 2022

@author: ljia
"""
import sys
import multiprocessing
import time
import numpy as np
import networkx as nx
from itertools import combinations
import multiprocessing
from multiprocessing import Pool

# from abc import ABC, abstractmethod
from sklearn.base import BaseEstimator  # , TransformerMixin
from sklearn.utils.validation import check_is_fitted  # check_X_y, check_array,
from sklearn.exceptions import NotFittedError

from gklearn.ged.model.distances import euclid_d
from gklearn.ged.util import pairwise_ged, get_nb_edit_operations
# from gklearn.utils import normalize_gram_matrix
from gklearn.utils import get_iters


# @TODO: it should be faster if creating a global env variable.
class GEDModel(BaseEstimator):  # , ABC):
	"""The graph edit distance model class compatible with `scikit-learn`.

	Attributes
	----------
	_graphs : list
		Stores the input graphs on fit input data.
		Default format of the list objects is `NetworkX` graphs.
		**We don't guarantee that the input graphs remain unchanged during the
		computation.**

	Notes
	-----
	This class uses the `gedlibpy` module to compute the graph edit distance.

	References
	----------
	https://ysig.github.io/GraKeL/0.1a8/_modules/grakel/kernels/kernel.html#Kernel.
	"""


	def __init__(
			self,
			ed_method='BIPARTITE',
			edit_cost_fun='CONSTANT',
			init_edit_cost_constants=[3, 3, 1, 3, 3, 1],
			optim_method='init',
			optim_options={'y_distance': euclid_d, 'mode': 'reg'},
			node_labels=[],
			edge_labels=[],
			parallel=None,
			n_jobs=None,
			chunksize=None,
			#			  normalize=True,
			copy_graphs=True,  # make sure it is a full deep copy. and faster!
			verbose=2
	):
		"""`__init__` for `GEDModel` object."""
		# @todo: the default settings of the parameters are different from those in the self.compute method.
		#		self._graphs = None
		self.ed_method = ed_method
		self.edit_cost_fun = edit_cost_fun
		self.init_edit_cost_constants = init_edit_cost_constants
		self.optim_method = optim_method
		self.optim_options = optim_options
		self.node_labels = node_labels
		self.edge_labels = edge_labels
		self.parallel = parallel
		self.n_jobs = (
			(multiprocessing.cpu_count() - 1) if n_jobs is None else n_jobs)
		self.chunksize = chunksize
		#		self.normalize = normalize
		self.copy_graphs = copy_graphs
		self.verbose = verbose


	#		self._run_time = 0
	#		self._gram_matrix = None
	#		self._gram_matrix_unnorm = None

	##########################################################################
	# The following is the 1st paradigm to compute GED distance matrix, which is
	# compatible with `scikit-learn`.
	##########################################################################

	def fit(self, X, y=None, **kwargs):
		"""Fit a graph dataset for a transformer.

		Parameters
		----------
		X : iterable
			DESCRIPTION.

		y : None, optional
			There is no need of a target in a transformer, yet the `scikit-learn`
			pipeline API requires this parameter.

		Returns
		-------
		object
			Returns self.

		"""
		#		self._is_tranformed = False

		# Clear any prior attributes stored on the estimator, # @todo: unless warm_start is used;
		self.clear_attributes()

		# Validate parameters for the transformer.
		self.validate_parameters()

		# Validate the input.
		self._graphs = self.validate_input(X)
		if y is not None:
			self._targets = y
		# self._targets = self.validate_input(y)

		# Compute edit cost constants.
		self.compute_edit_costs(**kwargs)

		#		self._X = X
		#		self._kernel = self._get_kernel_instance()

		# Return the transformer.
		return self


	def transform(
			self, X=None,
			return_dm_train=False,
			save_dm_test=False,
			return_dm_test=False,
			**kwargs
	):
		"""Compute the graph kernel matrix between given and fitted data.

		Parameters
		----------
		X : TYPE
			DESCRIPTION.

		Raises
		------
		ValueError
			DESCRIPTION.

		Returns
		-------
		None.

		"""
		# If `return_dm_train`, return the fitted GED distance matrix of training data.
		if return_dm_train:
			check_is_fitted(self, '_dm_train')
			self._is_transformed = True
			return self._dm_train  # @TODO: copy or not?

		if return_dm_test:
			check_is_fitted(self, '_dm_test')
			return self._dm_test  # @TODO: copy or not?

		# Check if method "fit" had been called.
		check_is_fitted(self, '_graphs')

		# Validate the input.
		Y = self.validate_input(X)

		# Transform: compute the graph kernel matrix.
		dis_matrix = self.compute_distance_matrix(Y, **kwargs)
		self._Y = Y

		# Self transform must appear before the diagonal call on normalization.
		self._is_transformed = True  # @TODO: When to set this to True? When return dm test?
		# 		if self.normalize:
		# 			X_diag, Y_diag = self.diagonals()
		# 			old_settings = np.seterr(invalid='raise') # Catch FloatingPointError: invalid value encountered in sqrt.
		# 			try:
		# 				kernel_matrix /= np.sqrt(np.outer(Y_diag, X_diag))
		# 			except:
		# 				raise
		# 			finally:
		# 				np.seterr(**old_settings)

		if save_dm_test:
			self._dm_test = dis_matrix
		# If the model is retransformed and the `save_dm_test` flag is not set,
		# then remove the previously computed dm_test to prevent conflicts.
		else:
			if hasattr(self, '_dm_test'):
				delattr(self, '_dm_test')

		return dis_matrix


	def fit_transform(
			self,
			X,
			y=None,
			save_dm_train=False,
			save_mm_train: bool = False,
			**kwargs):
		"""Fit and transform: compute GED distance matrix on the same data.

		Parameters
		----------
		X : list of graphs
			Input graphs.

		Returns
		-------
		dis_matrix : numpy array, shape = [len(X), len(X)]
			The distance matrix of X.

		"""
		self.fit(X, y, **kwargs)

		# Transform: compute Gram matrix.
		dis_matrix = self.compute_distance_matrix(**kwargs)

		#		# Normalize.
		#		if self.normalize:
		#			self._X_diag = np.diagonal(gram_matrix).copy()
		#			old_settings = np.seterr(invalid='raise') # Catch FloatingPointError: invalid value encountered in sqrt.
		#			try:
		#				gram_matrix /= np.sqrt(np.outer(self._X_diag, self._X_diag))
		#			except:
		#				raise
		#			finally:
		#				np.seterr(**old_settings)

		if save_mm_train or save_dm_train:
			self._dm_train = dis_matrix
		# If the model is refitted and the `save_dm_train` flag is not set, then
		# remove the previously computed dm_train to prevent conflicts.
		else:
			if hasattr(self, '_dm_train'):
				delattr(self, '_dm_train')

		return dis_matrix


	def get_params(self):
		pass


	def set_params(self):
		pass


	def clear_attributes(self):  # @todo: update
		#		if hasattr(self, '_X_diag'):
		#			delattr(self, '_X_diag')
		if hasattr(self, '_graphs'):
			delattr(self, '_graphs')
		if hasattr(self, '_Y'):
			delattr(self, '_Y')
		if hasattr(self, '_run_time'):
			delattr(self, '_run_time')
		if hasattr(self, '_test_run_time'):
			delattr(self, '_test_run_time')


	def validate_parameters(self):
		"""Validate all parameters for the transformer.

		Returns
		-------
		None.

		"""
		if self.parallel == False:
			self.parallel = None
		elif self.parallel == True:
			self.parallel = 'imap_unordered'
		if self.parallel is not None and self.parallel != 'imap_unordered':
			raise ValueError('Parallel mode is not set correctly.')

		if self.parallel == 'imap_unordered' and self.n_jobs is None:
			self.n_jobs = multiprocessing.cpu_count()


	def validate_input(self, X):
		"""Validate the given input and raise errors if it is invalid.

		Parameters
		----------
		X : list
			The input to check. Should be a list of graph.

		Raises
		------
		ValueError
			Raise if the input is not correct.

		Returns
		-------
		X : list
			The input. A list of graph.

		"""
		if X is None:
			raise ValueError('Please add graphs before computing.')
		elif not isinstance(X, list):
			raise ValueError('Cannot detect graphs. The input must be a list.')
		elif len(X) == 0:
			raise ValueError(
				'The graph list given is empty. No computation will be performed.'
			)

		return X


	def compute_distance_matrix(self, Y=None, **kwargs):
		"""Compute the distance matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) or the distance matrix for the fitted
		graphs (X / self._graphs).

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs. The default is None. If None distance is computed
			between X and itself.

		Returns
		-------
		dis_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed distance matrix.

		"""
		if Y is None:
			# Compute metric matrix for self._graphs (X).
			dis_matrix = self._compute_X_distance_matrix(**kwargs)
		#			self._gram_matrix_unnorm = np.copy(self._gram_matrix)

		else:
			# Compute metric matrix between Y and self._graphs (X).
			Y_copy = ([g.copy() for g in Y] if self.copy_graphs else Y)
			graphs_copy = (
				[g.copy() for g in self._graphs]
				if self.copy_graphs else self._graphs
			)

			start_time = time.time()

			if self.parallel == 'imap_unordered':
				dis_matrix = self._compute_distance_matrix_imap_unordered(Y)

			elif self.parallel is None:
				dis_matrix = self._compute_distance_matrix_series(
					Y_copy, graphs_copy, **kwargs
				)

			self._test_run_time = time.time() - start_time

			if self.verbose:
				print(
					'Distance matrix of size (%d, %d) built in %s seconds.'
					% (len(Y), len(self._graphs), self._test_run_time)
				)

		return dis_matrix


	def _compute_distance_matrix_series(self, X, Y, **kwargs):
		"""Compute the GED distance matrix between two sets of graphs (X and Y)
		without parallelization.

		Parameters
		----------
		X, Y : list of graphs
			The input graphs.

		Returns
		-------
		dis_matrix : numpy array, shape = [n_X, n_Y]
			The computed distance matrix.

		"""
		dis_matrix = np.zeros((len(X), len(Y)))

		for i_x, g_x in enumerate(X):
			for i_y, g_y in enumerate(Y):
				dis_matrix[i_x, i_y], _ = self.compute_ged(g_x, g_y, **kwargs)

		return dis_matrix


	def _compute_kernel_matrix_imap_unordered(self, Y):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) using imap unordered parallelization.

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		raise Exception('Parallelization for kernel matrix is not implemented.')


	def diagonals(self):
		"""Compute the kernel matrix diagonals of the fit/transformed data.

		Returns
		-------
        X_diag : numpy array
            The diagonal of the kernel matrix between the fitted data.
            This consists of each element calculated with itself.

        Y_diag : numpy array
            The diagonal of the kernel matrix, of the transform.
            This consists of each element calculated with itself.

		"""
		# Check if method "fit" had been called.
		check_is_fitted(self, ['_graphs'])

		# Check if the diagonals of X exist.
		try:
			check_is_fitted(self, ['_X_diag'])
		except NotFittedError:
			# Compute diagonals of X.
			self._X_diag = np.empty(shape=(len(self._graphs),))
			graphs = ([g.copy() for g in
			           self._graphs] if self.copy_graphs else self._graphs)
			for i, x in enumerate(graphs):
				self._X_diag[i] = self.pairwise_kernel(x, x)  # @todo: parallel?

		try:
			# If transform has happened, return both diagonals.
			check_is_fitted(self, ['_Y'])
			self._Y_diag = np.empty(shape=(len(self._Y),))
			Y = ([g.copy() for g in self._Y] if self.copy_graphs else self._Y)
			for (i, y) in enumerate(Y):
				self._Y_diag[i] = self.pairwise_kernel(y, y)  # @todo: parallel?

			return self._X_diag, self._Y_diag
		except NotFittedError:
			# Else just return both X_diag
			return self._X_diag


	#	@abstractmethod
	def pairwise_distance(self, x, y):
		"""Compute pairwise kernel between two graphs.

		Parameters
		----------
		x, y : NetworkX Graph.
			Graphs bewteen which the kernel is computed.

		Returns
		-------
		kernel: float
			The computed kernel.

#		Notes
#		-----
#		This method is abstract and must be implemented by a subclass.

		"""
		raise NotImplementedError(
			'Pairwise kernel computation is not implemented!'
		)


	def compute_edit_costs(self, Y=None, Y_targets=None, **kwargs):
		"""Compute edit cost constants. When optimizing method is `fiited`,
		apply Jia2021's metric learning method by using a given target graphs (Y)
		the fitted graphs (X / self._graphs).

		Parameters
		----------
		Y : TYPE, optional
			DESCRIPTION. The default is None.

		Returns
		-------
		None.

		"""
		# Get or compute.
		if self.optim_method == 'random':
			self._edit_cost_constants = np.random.rand(6)

		elif self.optim_method == 'init':
			self._edit_cost_constants = self.init_edit_cost_constants

		elif self.optim_method == 'expert':
			self._edit_cost_constants = [3, 3, 1, 3, 3, 1]

		elif self.optim_method == 'fitted':  # Jia2021 method
			# Get proper inputs.
			if Y is None:
				check_is_fitted(self, ['_graphs'])
				check_is_fitted(self, ['_targets'])
				graphs = ([g.copy() for g in
				           self._graphs] if self.copy_graphs else self._graphs)
				targets = self._targets
			else:
				graphs = ([g.copy() for g in Y] if self.copy_graphs else Y)
				targets = Y_targets

			# Get optimization options.
			node_labels = self.node_labels
			edge_labels = self.edge_labels
			unlabeled = (len(node_labels) == 0 and len(edge_labels) == 0)
			repeats = kwargs.get('repeats', 1)
			from gklearn.ged.model.optim_costs import compute_optimal_costs
			self._edit_cost_constants = compute_optimal_costs(
				graphs, targets,
				node_labels=node_labels, edge_labels=edge_labels,
				unlabeled=unlabeled,
				init_costs=self.init_edit_cost_constants,
				ed_method=self.ed_method,
				edit_cost_fun=self.edit_cost_fun,
				repeats=repeats,
				rescue_optim_failure=False,
				verbose=(self.verbose >= 2),
				**self.optim_options
			)


	##########################################################################
	# The following is the 2nd paradigm to compute kernel matrix. It is
	# simplified and not compatible with `scikit-learn`.
	##########################################################################

	#	def compute(self, *graphs, **kwargs):
	#		self.parallel = kwargs.get('parallel', 'imap_unordered')
	#		self.n_jobs = kwargs.get('n_jobs', multiprocessing.cpu_count())
	#		self.normalize = kwargs.get('normalize', True)
	#		self.verbose = kwargs.get('verbose', 2)
	#		self.copy_graphs = kwargs.get('copy_graphs', True)
	#		self.save_unnormed = kwargs.get('save_unnormed', True)
	#		self.validate_parameters()

	#		# If the inputs is a list of graphs.
	#		if len(graphs) == 1:
	#			if not isinstance(graphs[0], list):
	#				raise Exception('Cannot detect graphs.')
	#			elif len(graphs[0]) == 0:
	#				raise Exception('The graph list given is empty. No computation was performed.')
	#			else:
	#				if self.copy_graphs:
	#					self._graphs = [g.copy() for g in graphs[0]] # @todo: might be very slow.
	#				else:
	#					self._graphs = graphs
	#				self._gram_matrix = self._compute_gram_matrix()

	#				if self.save_unnormed:
	#					self._gram_matrix_unnorm = np.copy(self._gram_matrix)
	#				if self.normalize:
	#					self._gram_matrix = normalize_gram_matrix(self._gram_matrix)
	#				return self._gram_matrix, self._run_time

	#		elif len(graphs) == 2:
	#			# If the inputs are two graphs.
	#			if self.is_graph(graphs[0]) and self.is_graph(graphs[1]):
	#				if self.copy_graphs:
	#					G0, G1 = graphs[0].copy(), graphs[1].copy()
	#				else:
	#					G0, G1 = graphs[0], graphs[1]
	#				kernel = self._compute_single_kernel(G0, G1)
	#				return kernel, self._run_time

	#			# If the inputs are a graph and a list of graphs.
	#			elif self.is_graph(graphs[0]) and isinstance(graphs[1], list):
	#				if self.copy_graphs:
	#					g1 = graphs[0].copy()
	#					g_list = [g.copy() for g in graphs[1]]
	#					kernel_list = self._compute_kernel_list(g1, g_list)
	#				else:
	#					kernel_list = self._compute_kernel_list(graphs[0], graphs[1])
	#				return kernel_list, self._run_time

	#			elif isinstance(graphs[0], list) and self.is_graph(graphs[1]):
	#				if self.copy_graphs:
	#					g1 = graphs[1].copy()
	#					g_list = [g.copy() for g in graphs[0]]
	#					kernel_list = self._compute_kernel_list(g1, g_list)
	#				else:
	#					kernel_list = self._compute_kernel_list(graphs[1], graphs[0])
	#				return kernel_list, self._run_time

	#			else:
	#				raise Exception('Cannot detect graphs.')

	#		elif len(graphs) == 0 and self._graphs is None:
	#			raise Exception('Please add graphs before computing.')

	#		else:
	#			raise Exception('Cannot detect graphs.')

	#	def normalize_gm(self, gram_matrix):
	#		import warnings
	#		warnings.warn('gklearn.kernels.graph_kernel.normalize_gm will be deprecated, use gklearn.utils.normalize_gram_matrix instead', DeprecationWarning)

	#		diag = gram_matrix.diagonal().copy()
	#		for i in range(len(gram_matrix)):
	#			for j in range(i, len(gram_matrix)):
	#				gram_matrix[i][j] /= np.sqrt(diag[i] * diag[j])
	#				gram_matrix[j][i] = gram_matrix[i][j]
	#		return gram_matrix

	#	def compute_distance_matrix(self):
	#		if self._gram_matrix is None:
	#			raise Exception('Please compute the Gram matrix before computing distance matrix.')
	#		dis_mat = np.empty((len(self._gram_matrix), len(self._gram_matrix)))
	#		for i in range(len(self._gram_matrix)):
	#			for j in range(i, len(self._gram_matrix)):
	#				dis = self._gram_matrix[i, i] + self._gram_matrix[j, j] - 2 * self._gram_matrix[i, j]
	#				if dis < 0:
	#					if dis > -1e-10:
	#						dis = 0
	#					else:
	#						raise ValueError('The distance is negative.')
	#				dis_mat[i, j] = np.sqrt(dis)
	#				dis_mat[j, i] = dis_mat[i, j]
	#		dis_max = np.max(np.max(dis_mat))
	#		dis_min = np.min(np.min(dis_mat[dis_mat != 0]))
	#		dis_mean = np.mean(np.mean(dis_mat))
	#		return dis_mat, dis_max, dis_min, dis_mean

	def _compute_X_distance_matrix(self, **kwargs):
		graphs = ([g.copy() for g in
		           self._graphs] if self.copy_graphs else self._graphs)

		start_time = time.time()

		if self.parallel == 'imap_unordered':
			dis_matrix = self._compute_X_dm_imap_unordered(graphs, **kwargs)
		elif self.parallel is None:
			dis_matrix = self._compute_X_dm_series(graphs, **kwargs)
		else:
			raise Exception('Parallel mode is not set correctly.')

		self._run_time = time.time() - start_time

		if self.verbose:
			print(
				'Distance matrix of size %d built in %s seconds.'
				% (len(self._graphs), self._run_time)
			)

		return dis_matrix


	def _compute_X_dm_series(self, graphs, **kwargs):
		n = len(graphs)
		dis_matrix = np.zeros((n, n))

		iterator = combinations(range(n), 2)
		len_itr = int(n * (n - 1) / 2)
		if self.verbose:
			print('Graphs in total: %d.' % len(graphs))
			print('The total # of pairs is %d.' % len_itr)
		for i, j in get_iters(
				iterator, desc='Computing distance matrix',
				file=sys.stdout, verbose=(self.verbose >= 2), length=len_itr
		):
			g1, g2 = graphs[i], graphs[j]
			dis_matrix[i, j], _ = self.compute_ged(g1, g2, **kwargs)
			dis_matrix[j, i] = dis_matrix[i, j]
		return dis_matrix


	def _compute_X_dm_imap_unordered(self, graphs, **kwargs):
		"""Compute GED distance matrix in parallel using imap_unordered.
		"""
		# This is very slow, maybe because of the Cython is involved.
		from gklearn.utils.parallel import parallel_ged_mat
		n = len(graphs)
		dis_matrix = np.zeros((n, n))
		if self.verbose:
			print('Graphs in total: %d.' % len(graphs))
			print('The total # of pairs is %d.' % int(n * (n + 1) / 2))

		do_fun = self._wrapper_compute_ged
		parallel_ged_mat(
			do_fun, dis_matrix, graphs, init_worker=_init_worker_ged_mat,
			glbv=(graphs,), n_jobs=self.n_jobs, verbose=self.verbose
		)


	def _wrapper_compute_ged(self, itr):
		i = itr[0]
		j = itr[1]
		# @TODO: repeats are not considered here.
		dis, _ = self.compute_ged(G_gn[i], G_gn[j])
		return i, j, dis


	# # imap_unordered returns an iterator of the results in the order
	# # in which the function calls are started.
	# # Note that imap_unordered may end up consuming all of the
	# # available memory if the iterable is too large.
	# n = len(graphs)
	# dis_matrix = np.zeros((n, n))
	# iterator = combinations(range(n), 2)
	# len_itr = int(n * (n + 1) / 2)
	# pool = Pool(processes=self.n_jobs)
	# for i, j in get_iters(
	# 		iterator, desc='Computing distance matrix',
	# 		file=sys.stdout, verbose=(self.verbose >= 2), length=len_itr
	# ):
	# 	g1, g2 = graphs[i], graphs[j]
	# 	dis_matrix[i, j], _ = pool.apply_async(
	# 		self.compute_ged, (g1, g2)
	# 	).get()
	# 	dis_matrix[j, i] = dis_matrix[i, j]
	# pool.close()
	# return dis_matrix

	def compute_ged(self, Gi, Gj, **kwargs):
		"""
		Compute GED between two graph according to edit_cost.
		"""
		ged_options = {
			'edit_cost': self.edit_cost_fun,
			'method': self.ed_method,
			'edit_cost_constants': self._edit_cost_constants
		}
		repeats = kwargs.get('repeats', 1)
		dis, pi_forward, pi_backward = pairwise_ged(
			Gi, Gj, ged_options, repeats=repeats
		)
		# @TODO: Better to have a if here.
		# if self.compute_n_eo:
		# 	n_eo_tmp = get_nb_edit_operations(
		# 		Gi, Gj, pi_forward, pi_backward,
		# 		edit_cost=self.edit_cost_fun,
		# 		node_labels=self.node_labels, edge_labels=self.edge_labels
		# 	)
		# else:
		# 	n_eo_tmp = None
		# return dis, n_eo_tmp
		return dis, None


	# 	def _compute_kernel_list(self, g1, g_list):
	# 		start_time = time.time()

	# 		if self.parallel == 'imap_unordered':
	# 			kernel_list = self._compute_kernel_list_imap_unordered(g1, g_list)
	# 		elif self.parallel is None:
	# 			kernel_list = self._compute_kernel_list_series(g1, g_list)
	# 		else:
	# 			raise Exception('Parallel mode is not set correctly.')

	# 		self._run_time = time.time() - start_time
	# 		if self.verbose:
	# 			print('Graph kernel bewteen a graph and a list of %d graphs built in %s seconds.'
	# 			  % (len(g_list), self._run_time))

	# 		return kernel_list

	# 	def _compute_kernel_list_series(self, g1, g_list):
	# 		pass

	# 	def _compute_kernel_list_imap_unordered(self, g1, g_list):
	# 		pass

	# 	def _compute_single_kernel(self, g1, g2):
	# 		start_time = time.time()

	# 		kernel = self._compute_single_kernel_series(g1, g2)

	# 		self._run_time = time.time() - start_time
	# 		if self.verbose:
	# 			print('Graph kernel bewteen two graphs built in %s seconds.' % (self._run_time))

	# 		return kernel

	# 	def _compute_single_kernel_series(self, g1, g2):
	# 		pass

	def is_graph(self, graph):
		if isinstance(graph, nx.Graph):
			return True
		if isinstance(graph, nx.DiGraph):
			return True
		if isinstance(graph, nx.MultiGraph):
			return True
		if isinstance(graph, nx.MultiDiGraph):
			return True
		return False


	def __repr__(self):
		return (
			f"{self.__class__.__name__}("
			f"optim_method={self.optim_method}, "
			f"ed_method={self.ed_method}, "
			f"edit_cost_fun={self.edit_cost_fun}, "
			f"node_labels={self.node_labels}, "
			f"edge_labels={self.edge_labels}, "
			f"optim_options={self.optim_options}, "
			f"init_edit_cost_constants={self.init_edit_cost_constants}, "
			f"copy_graphs={self.copy_graphs}, "
			f"parallel={self.parallel}, "
			f"n_jobs={self.n_jobs}, "
			f"verbose={self.verbose}, "
			f"normalize={self.normalize}, "
			f"run_time={self.run_time}"
			f")"
		)


	@property
	def graphs(self):
		return self._graphs


	#	@property
	#	def parallel(self):
	#		return self.parallel

	#	@property
	#	def n_jobs(self):
	#		return self.n_jobs

	#	@property
	#	def verbose(self):
	#		return self.verbose

	#	@property
	#	def normalize(self):
	#		return self.normalize

	@property
	def run_time(self):
		return self._run_time


	@property
	def test_run_time(self):
		return self._test_run_time


	@property
	def dis_matrix(self):
		return self._dm_train

	@dis_matrix.setter
	def dis_matrix(self, value):
		self._dm_train = value


	@property
	def metric_matrix(self):
		return self._dm_train


	@metric_matrix.setter
	def metric_matrix(self, value):
		self._dm_train = value


	@property
	def edit_cost_constants(self):
		return self._edit_cost_constants


# 	@property
# 	def gram_matrix_unnorm(self):
# 		return self._gram_matrix_unnorm

# 	@gram_matrix_unnorm.setter
# 	def gram_matrix_unnorm(self, value):
# 		self._gram_matrix_unnorm = value

	@property
	def n_pairs(self):
		"""
		The number of pairs of graphs between which the GEDs are computed.
		"""
		try:
			check_is_fitted(self, '_dm_train')
			return len(self._dm_train) * (len(self._dm_train) - 1) / 2
		except NotFittedError:
			return None


def _init_worker_ged_mat(gn_toshare):
	global G_gn
	G_gn = gn_toshare



from gklearn.ged.env.common_types import Options, OptionsStringMap, AlgorithmState
from gklearn.ged.env.ged_data import GEDData
from gklearn.ged.env.ged_env import GEDEnv
from gklearn.ged.env.node_map import NodeMap


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 30 10:48:49 2019

Trie (prefix tree)
@author: ljia
@references:
`NLP: Build a Trie Data structure from scratch with python <https://viblo.asia/p/nlp-build-a-trie-data-structure-from-scratch-with-python-3P0lPzroKox>`__, 2019.1
"""

import pickle
import json

""" Trie class
"""
class Trie:
    """
    """
    # init Trie class
    def __init__(self):
        self.root = self.getNode()

    def getNode(self):
        return {"isEndOfWord": False, "children": {}}

    def insertWord(self, word):
        current = self.root
        for ch in word:

            if ch in current["children"]:
                node = current["children"][ch]
            else:
                node = self.getNode()
                current["children"][ch] = node

            current = node
        current["isEndOfWord"] = True
        if 'count' in current:
            current['count'] += 1
        else:
            current['count'] = 1

    def searchWord(self, word):
        current = self.root
        for ch in word:
            if ch not in current["children"]:
                return 0
            node = current["children"][ch]

            current = node
        if 'count' in current:
            return current["count"]
        else:
            return 0

    def searchWordPrefix(self, word):
        current = self.root
        for ch in word:
            if not current["children"].has_key(ch):
                return False
            node = current["children"][ch]

            current = node
        # return True if children contain keys and values
        return bool(current["children"])

    def deleteWord(self, word):
        self._delete(self.root, word, 0)

    def _delete(self, current, word, index):
        if(index == len(word)):
            if not current["isEndOfWord"]:
                return False
            current["isEndOfWord"] = False
            return len(current["children"].keys()) == 0

        ch = word[index]
        if not current["children"].has_key(ch):
            return False
        node = current["children"][ch]

        should_delete_current_node = self._delete(node, word, index + 1)

        if should_delete_current_node:
            current["children"].pop(ch)
            return len(current["children"].keys()) == 0

        return False

    def save_to_pickle(self, file_name):
        f = open(file_name + ".pkl", "wb")
        pickle.dump(self.root, f)
        f.close()

    def load_from_pickle(self, file_name):
        f = open(file_name + ".pkl", "rb")
        self.root = pickle.load(f)
        f.close()
        
    def to_json(self):
        return json.dump(self.root)

    def save_to_json(self, file_name):
        json_data = json.dumps(self.root)
        f = open(file_name + ".json", "w")
        f.write(json_data)
        f.close()

    def load_from_json(self, file_name):
        json_file = open(file_name + ".json", "r")
        self.root = json.load(json_file)
        json_file.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 18 11:21:31 2020

@author: ljia

@references:

	[1] Thomas Gärtner, Peter Flach, and Stefan Wrobel. On graph kernels:
	Hardness results and efficient alternatives. Learning Theory and Kernel
	Machines, pages 129–143, 2003.
"""

import sys
from gklearn.utils import get_iters
import numpy as np
import networkx as nx
from gklearn.utils import SpecialLabel
from gklearn.utils.parallel import parallel_gm, parallel_me
from gklearn.utils.utils import direct_product_graph
from gklearn.kernels import GraphKernel


class CommonWalk(GraphKernel):

	def __init__(self, **kwargs):
		GraphKernel.__init__(self)
		self._node_labels = kwargs.get('node_labels', [])
		self._edge_labels = kwargs.get('edge_labels', [])
		self._weight = kwargs.get('weight', 1)
		self._compute_method = kwargs.get('compute_method', None)
		self._ds_infos = kwargs.get('ds_infos', {})
		self._compute_method = self._compute_method.lower()


	def _compute_gm_series(self, graphs):
		self._check_graphs(graphs)
		self._add_dummy_labels(graphs)
		if not self._ds_infos['directed']:  #  convert
			graphs = [G.to_directed() for G in graphs]

		# compute Gram matrix.
		gram_matrix = np.zeros((len(graphs), len(graphs)))

		from itertools import combinations_with_replacement
		itr = combinations_with_replacement(range(0, len(graphs)), 2)
		len_itr = int(len(graphs) * (len(graphs) + 1) / 2)
		iterator = get_iters(itr, desc='Computing kernels', file=sys.stdout,
					length=len_itr, verbose=(self.verbose >= 2))

		# direct product graph method - exponential
		if self._compute_method == 'exp':
			for i, j in iterator:
				kernel = self._kernel_do_exp(graphs[i], graphs[j], self._weight)
				gram_matrix[i][j] = kernel
				gram_matrix[j][i] = kernel
		# direct product graph method - geometric
		elif self._compute_method == 'geo':
			for i, j in iterator:
				kernel = self._kernel_do_geo(graphs[i], graphs[j], self._weight)
				gram_matrix[i][j] = kernel
				gram_matrix[j][i] = kernel

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		self._check_graphs(self._graphs)
		self._add_dummy_labels(self._graphs)
		if not self._ds_infos['directed']:  #  convert
			self._graphs = [G.to_directed() for G in self._graphs]

		# compute Gram matrix.
		gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))

# 		def init_worker(gn_toshare):
# 			global G_gn
# 			G_gn = gn_toshare

		# direct product graph method - exponential
		if self._compute_method == 'exp':
			do_fun = self._wrapper_kernel_do_exp
		# direct product graph method - geometric
		elif self._compute_method == 'geo':
			do_fun = self._wrapper_kernel_do_geo

		parallel_gm(do_fun, gram_matrix, self._graphs, init_worker=_init_worker_gm,
			  glbv=(self._graphs,), n_jobs=self.n_jobs, verbose=self.verbose)

		return gram_matrix


	def _compute_kernel_list_series(self, g1, g_list):
		self._check_graphs(g_list + [g1])
		self._add_dummy_labels(g_list + [g1])
		if not self._ds_infos['directed']:  #  convert
			g1 = g1.to_directed()
			g_list = [G.to_directed() for G in g_list]

		# compute kernel list.
		kernel_list = [None] * len(g_list)
		if self.verbose >= 2:
			iterator = get_iters(range(len(g_list)), desc='Computing kernels',
						 file=sys.stdout, length=len(g_list), verbose=(self.verbose >= 2))
		else:
			iterator = range(len(g_list))

		# direct product graph method - exponential
		if self._compute_method == 'exp':
			for i in iterator:
				kernel = self._kernel_do_exp(g1, g_list[i], self._weight)
				kernel_list[i] = kernel
		# direct product graph method - geometric
		elif self._compute_method == 'geo':
			for i in iterator:
				kernel = self._kernel_do_geo(g1, g_list[i], self._weight)
				kernel_list[i] = kernel

		return kernel_list


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._check_graphs(g_list + [g1])
		self._add_dummy_labels(g_list + [g1])
		if not self._ds_infos['directed']:  #  convert
			g1 = g1.to_directed()
			g_list = [G.to_directed() for G in g_list]

		# compute kernel list.
		kernel_list = [None] * len(g_list)

# 		def init_worker(g1_toshare, g_list_toshare):
# 			global G_g1, G_g_list
# 			G_g1 = g1_toshare
# 			G_g_list = g_list_toshare

		# direct product graph method - exponential
		if self._compute_method == 'exp':
			do_fun = self._wrapper_kernel_list_do_exp
		# direct product graph method - geometric
		elif self._compute_method == 'geo':
			do_fun = self._wrapper_kernel_list_do_geo

		def func_assign(result, var_to_assign):
			var_to_assign[result[0]] = result[1]
		itr = range(len(g_list))
		len_itr = len(g_list)
		parallel_me(do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
			init_worker=_init_worker_list, glbv=(g1, g_list), method='imap_unordered',
			n_jobs=self.n_jobs, itr_desc='Computing kernels', verbose=self.verbose)

		return kernel_list


	def _wrapper_kernel_list_do_exp(self, itr):
		return itr, self._kernel_do_exp(G_g1, G_g_list[itr], self._weight)


	def _wrapper_kernel_list_do_geo(self, itr):
		return itr, self._kernel_do_geo(G_g1, G_g_list[itr], self._weight)


	def _compute_single_kernel_series(self, g1, g2):
		self._check_graphs([g1] + [g2])
		self._add_dummy_labels([g1] + [g2])
		if not self._ds_infos['directed']:  #  convert
			g1 = g1.to_directed()
			g2 = g2.to_directed()

		# direct product graph method - exponential
		if self._compute_method == 'exp':
			kernel = self._kernel_do_exp(g1, g2, self._weight)
		# direct product graph method - geometric
		elif self._compute_method == 'geo':
			kernel = self._kernel_do_geo(g1, g2, self._weight)

		return kernel


	def _kernel_do_exp(self, g1, g2, beta):
		"""Compute common walk graph kernel between 2 graphs using exponential
		series.

		Parameters
		----------
		g1, g2 : NetworkX graphs
			Graphs between which the kernels are computed.
		beta : integer
			Weight.

		Return
		------
		kernel : float
			The common walk Kernel between 2 graphs.
		"""
		# get tensor product / direct product
		gp = direct_product_graph(g1, g2, self._node_labels, self._edge_labels)
		# return 0 if the direct product graph have no more than 1 node.
		if nx.number_of_nodes(gp) < 2:
			return 0
		A = nx.adjacency_matrix(gp).todense()

		ew, ev = np.linalg.eig(A)
# 		# remove imaginary part if possible.
# 		# @todo: don't know if it is necessary.
# 		for i in range(len(ew)):
# 			if np.abs(ew[i].imag) < 1e-9:
# 				ew[i] = ew[i].real
# 		for i in range(ev.shape[0]):
# 			for j in range(ev.shape[1]):
# 				if np.abs(ev[i, j].imag) < 1e-9:
# 					ev[i, j] = ev[i, j].real

		D = np.zeros((len(ew), len(ew)), dtype=complex) # @todo: use complex?
		for i in range(len(ew)):
			D[i][i] = np.exp(beta * ew[i])

		exp_D = ev * D * ev.T
		kernel = exp_D.sum()
		if (kernel.real == 0 and np.abs(kernel.imag) < 1e-9) or np.abs(kernel.imag / kernel.real) < 1e-9:
			kernel = kernel.real

		return kernel


	def _wrapper_kernel_do_exp(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._kernel_do_exp(G_gn[i], G_gn[j], self._weight)


	def _kernel_do_geo(self, g1, g2, gamma):
		"""Compute common walk graph kernel between 2 graphs using geometric
		series.

		Parameters
		----------
		g1, g2 : NetworkX graphs
			Graphs between which the kernels are computed.
		gamma : integer
			Weight.

		Return
		------
		kernel : float
			The common walk Kernel between 2 graphs.
		"""
		# get tensor product / direct product
		gp = direct_product_graph(g1, g2, self._node_labels, self._edge_labels)
		# return 0 if the direct product graph have no more than 1 node.
		if nx.number_of_nodes(gp) < 2:
			return 0
		A = nx.adjacency_matrix(gp).todense()
		mat = np.identity(len(A)) - gamma * A
	#	try:
		return np.linalg.inv(mat).sum()
	#	except np.linalg.LinAlgError:
	#		return np.nan


	def _wrapper_kernel_do_geo(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._kernel_do_geo(G_gn[i], G_gn[j], self._weight)


	def _check_graphs(self, Gn):
		for g in Gn:
			if nx.number_of_nodes(g) == 1:
				raise Exception('Graphs must contain more than 1 nodes to construct adjacency matrices.')


	def _add_dummy_labels(self, Gn):
		if len(self._node_labels) == 0 or (len(self._node_labels) == 1 and self._node_labels[0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_node_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self._node_labels = [SpecialLabel.DUMMY]
		if len(self._edge_labels) == 0 or (len(self._edge_labels) == 1 and self._edge_labels[0] == SpecialLabel.DUMMY):
			for i in range(len(Gn)):
				nx.set_edge_attributes(Gn[i], '0', SpecialLabel.DUMMY)
			self._edge_labels = [SpecialLabel.DUMMY]


def _init_worker_gm(gn_toshare):
	global G_gn
	G_gn = gn_toshare


def _init_worker_list(g1_toshare, g_list_toshare):
	global G_g1, G_g_list
	G_g1 = g1_toshare
	G_g_list = g_list_toshare


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 11 11:39:46 2018
Parallel aid functions.
@author: ljia
"""
import multiprocessing
import sys
from multiprocessing import Pool

from gklearn.utils import get_iters


def parallel_me(
		func, func_assign, var_to_assign, itr, len_itr=None, init_worker=None,
		glbv=None, method=None, n_jobs=None, chunksize=None, itr_desc='',
		verbose=True
):
	'''
	'''
	if method == 'imap_unordered':
		if glbv:  # global varibles required.
			#			def init_worker(v_share):
			#				global G_var
			#				G_var = v_share
			if n_jobs == None:
				n_jobs = multiprocessing.cpu_count()
			with Pool(
					processes=n_jobs, initializer=init_worker,
					initargs=glbv
			) as pool:
				if chunksize is None:
					if len_itr < 100 * n_jobs:
						chunksize = int(len_itr / n_jobs) + 1
					else:
						chunksize = 100

				iterator = get_iters(
					pool.imap_unordered(func, itr, chunksize),
					desc=itr_desc, file=sys.stdout, length=len_itr,
					verbose=(verbose >= 2)
				)
				for result in iterator:
					func_assign(result, var_to_assign)
			pool.close()
			pool.join()
		else:
			if n_jobs == None:
				n_jobs = multiprocessing.cpu_count()
			with Pool(processes=n_jobs) as pool:
				if chunksize is None:
					if len_itr < 100 * n_jobs:
						chunksize = int(len_itr / n_jobs) + 1
					else:
						chunksize = 100
				iterator = get_iters(
					pool.imap_unordered(func, itr, chunksize),
					desc=itr_desc, file=sys.stdout, length=len_itr,
					verbose=(verbose >= 2)
				)
				for result in iterator:
					func_assign(result, var_to_assign)
			pool.close()
			pool.join()


def parallel_gm(
		func, Kmatrix, Gn, init_worker=None, glbv=None,
		method='imap_unordered', n_jobs=None, chunksize=None,
		verbose=True
):  # @todo: Gn seems not necessary.
	from itertools import combinations_with_replacement

	def func_assign(result, var_to_assign):
		var_to_assign[result[0]][result[1]] = result[2]
		var_to_assign[result[1]][result[0]] = result[2]

	itr = combinations_with_replacement(range(0, len(Gn)), 2)
	len_itr = int(len(Gn) * (len(Gn) + 1) / 2)
	parallel_me(
		func, func_assign, Kmatrix, itr, len_itr=len_itr,
		init_worker=init_worker, glbv=glbv, method=method, n_jobs=n_jobs,
		chunksize=chunksize, itr_desc='Computing kernels', verbose=verbose
	)


def parallel_ged_mat(
		func, ged_mat, Gn, init_worker=None, glbv=None,
		method='imap_unordered', n_jobs=None, chunksize=None,
		verbose=True
):
	"""Parallel computing graph edit distance matrix.

	Notes
	-----
	This is equivalent to the function `parallel_gm`.
	"""
	# @todo: Gn seems not necessary.
	from itertools import combinations_with_replacement

	def func_assign(result, var_to_assign):
		var_to_assign[result[0]][result[1]] = result[2]
		var_to_assign[result[1]][result[0]] = result[2]

	itr = combinations_with_replacement(range(0, len(Gn)), 2)
	len_itr = int(len(Gn) * (len(Gn) + 1) / 2)
	parallel_me(
		func, func_assign, ged_mat, itr, len_itr=len_itr,
		init_worker=init_worker, glbv=glbv, method=method, n_jobs=n_jobs,
		chunksize=chunksize, itr_desc='Computing GED matrix', verbose=verbose
	)



# -*-coding:utf-8 -*-
"""gklearn - graph kernels module
"""

# info
__version__ = "0.1"
__author__ = "Linlin Jia"
__date__ = "November 2018"


from gklearn.kernels.graph_kernel import GraphKernel
from gklearn.kernels.common_walk import CommonWalk
from gklearn.kernels.marginalized import Marginalized
from gklearn.kernels.random_walk_meta import RandomWalkMeta
from gklearn.kernels.sylvester_equation import SylvesterEquation
from gklearn.kernels.conjugate_gradient import ConjugateGradient
from gklearn.kernels.fixed_point import FixedPoint
from gklearn.kernels.spectral_decomposition import SpectralDecomposition
from gklearn.kernels.random_walk import RandomWalk
from gklearn.kernels.shortest_path import ShortestPath
from gklearn.kernels.structural_sp import StructuralSP
from gklearn.kernels.path_up_to_h import PathUpToH
from gklearn.kernels.treelet import Treelet
from gklearn.kernels.weisfeiler_lehman import WeisfeilerLehman, WLSubtree

from gklearn.kernels.metadata import GRAPH_KERNELS, list_of_graph_kernels

# old version.
from gklearn.kernels.commonWalkKernel import commonwalkkernel
from gklearn.kernels.marginalizedKernel import marginalizedkernel
from gklearn.kernels.randomWalkKernel import randomwalkkernel
from gklearn.kernels.spKernel import spkernel
from gklearn.kernels.structuralspKernel import structuralspkernel
from gklearn.kernels.untilHPathKernel import untilhpathkernel
from gklearn.kernels.treeletKernel import treeletkernel
from gklearn.kernels.weisfeilerLehmanKernel import weisfeilerlehmankernel



# -*-coding:utf-8 -*-
"""gklearn - preimage module

Implements pre-image method for graphs.

"""

# info
__version__ = "0.1"
__author__ = "Linlin Jia"
__date__ = "March 2020"

from gklearn.preimage.preimage_generator import PreimageGenerator
from gklearn.preimage.random_preimage_generator import RandomPreimageGenerator
from gklearn.preimage.median_preimage_generator import MedianPreimageGenerator
from gklearn.preimage.median_preimage_generator_py import MedianPreimageGeneratorPy
from gklearn.preimage.median_preimage_generator_cml import MedianPreimageGeneratorCML
from gklearn.preimage.kernel_knn_cv import kernel_knn_cv
from gklearn.preimage.generate_random_preimages_by_class import generate_random_preimages_by_class



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov  8 14:21:25 2019

@author: ljia
"""

import sys
import time

class Logger(object):
    def __init__(self):
        self.terminal = sys.stdout
        self.log = open("log." + str(time.time()) + ".log", "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)  

    def flush(self):
        #this flush method is needed for python 3 compatibility.
        #this handles the flush command by doing nothing.
        #you might want to specify some extra behavior here.
        pass    

sys.stdout = Logger()


from distutils.extension import Extension
from distutils.core import setup
from Cython.Build import cythonize

import os
import sys
import shutil
import subprocess
import platform

import numpy
import importlib


def parse_args():
	# @TODO: This may not work with bdist_wheel/PyPI installation.
	import argparse
	parser = argparse.ArgumentParser()

	parser.add_argument(
		'--use-existing-gedlib',
		dest='use_existing_gedlib',
		type=str,
		choices=['true', 'false'],
		default='false',
		help='Whether to use an existing GEDLIB C++ library. This will avoid '
		     'downloading the library from GitHub and extracting it. If no library '
		     'is found, the installation will stop and an error will be raised. '
		     'Does not have effect when `--build-gedlibpy` is set to `false`. It may help '
		     'when you have problem accessing GitHub or it takes too long time to '
		     'extract the file (which is my case on CentOS (I hate this system!)). '
	)

	parser.add_argument(
		'--build-gedlib',
		dest='build_gedlib',
		type=str,
		choices=['true', 'false'],
		default='true',
		help='Whether to build GEDLIB C++.'
	)

	parser.add_argument(
		'--develop-mode',
		dest='develop_mode',
		type=str,
		choices=['true', 'false'],
		default='true',
		help='Whether in development mode. If true, the include files in the `gedlibpy` module will be deleted after installation.'
	)

	args, unknown = parser.parse_known_args()
	return args


def install(package):
	try:
		importlib.import_module(package)
	except ImportError:
		import subprocess
		cur_python = sys.executable
		subprocess.call([cur_python, '-m', 'pip', 'install', package])


def get_gedlib():
	install('tqdm')

	import urllib.request
	from tqdm import tqdm
	import zipfile

	print()
	# Delete previous GEDLIB installation if exists:
	path_gedlib = os.path.join(os.getcwd(), 'include/gedlib-master/')
	if os.path.exists(path_gedlib):
		print('Deleting previous GEDLIB installation...')
		shutil.rmtree(path_gedlib)
		print('Done!')
		print()
	else:
		pass
		print('No previous installation of GEDLIB.')

	print('Downloading and unpacking GEDLIB from GitHub...')

	url = 'https://github.com/jajupmochi/gedlib/archive/refs/heads/master.zip'
	filename = 'gedlib-master.zip'

	with tqdm(
			unit='B', unit_scale=True, unit_divisor=1024, miniters=1,
			desc=filename
	) as t:
		urllib.request.urlretrieve(
			url, filename=filename,
			reporthook=lambda blocknum, blocksize, totalsize: t.update(
				blocknum * blocksize - t.n
			)
		)

	with zipfile.ZipFile(filename, 'r') as zip_ref:
		for member in tqdm(zip_ref.infolist(), desc='Extracting'):
			try:
				zip_ref.extract(member, path='include/')
			except zipfile.error as e:
				pass

	# Delete the .zip file:
	os.remove(filename)
	print('The .zip file of GEDLIB is deleted.')

	print('Done!')
	print()


def check_gedlib():
	print()
	print('Checking if GEDLIB is already downloaded...')
	path_gedlib = os.path.join(os.getcwd(), 'include/gedlib-master/')
	if not os.path.exists(path_gedlib):
		raise ModuleError(
			'The C++ library `GEDLIB` is not found, which is required to build the '
			'`gedlibpy` module. You have to install several ways to fix this error:'
			'\n-- 1. Download GEDLIB from '
			'`https://github.com/jajupmochi/gedlib/archive/refs/heads/master.zip`, '
			'and extract it into `include/gedlib-master/`.'
			'\n-- 2. Set `use-existing-gedlib` to `false` or the default value.'
			'\n-- 3. Set `--build-gedlibpy` to `false`. Notice, it '
			'is possible that the module would not work when it is incompatible '
			'with the system (libraries).'
		)
	print('Done!')
	print()


def install_gedlib():
	"""
	Notes
	-----
	The followings are done during this installation:
		- Build files in `ext/fann.2.2.0/lib/`;
		- Clean obj files in `ext/nomad.3.8.1/ext/sgtelib/src/`;
		- Build files in `ext/nomad.3.8.1/ext/sgtelib/lib/`;
		- Generate executable files in `ext/nomad.3.8.1/ext/sgtelib/bin/`;
		- Build files in `ext/nomad.3.8.1/lib/`;
		- Generate `ext/libsvm.3.22/libsvm.so`;
	"""
	print()
	print('Installing GEDLIB...')
	import subprocess
	cur_python = sys.executable
	subprocess.call([cur_python, '--version'])
	subprocess.call(['which', cur_python])
	gedlib_dir = 'include/gedlib-master/'
	subprocess.call(
		[cur_python, 'install.py'], cwd=gedlib_dir
	)
	print()


def copy_gedlib_includes_and_libs():
	def copy_paste_with_logs(src_path, dst_path):
		"""
		Authors
		-------
		Linlin Jia, ChatGPT 3.5 (2023.04.27)
		"""
		cur_dir = os.path.dirname(os.path.abspath(__file__))
		src_path = os.path.join(cur_dir, src_path)
		dst_path = os.path.join(cur_dir, dst_path)

		# if the source is a file, copy it and print the copied file name
		if os.path.isfile(src_path):
			shutil.copy(src_path, dst_path)
			print(f"Copied file {src_path} to {dst_path}.")

		# if the source is a directory, copy it and print the copied file names
		if os.path.isdir(src_path):
			shutil.copytree(src_path, dst_path, dirs_exist_ok=True)
			print(f"Copied directory {src_path} to {dst_path}.")


	print()
	print(
		'Copying and pasting (generated) includes and libs files of GEDLIB into '
		'proper locations to build gedlibpy...'
	)
	# Copy-paste includes:
	print('Copying includes...')
	copy_paste_with_logs(
		'include/gedlib-master/ext/boost.1.69.0/', 'include/boost.1.69.0/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/eigen.3.3.4/Eigen/',
		'include/eigen.3.3.4/Eigen/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/nomad.3.8.1/src/', 'include/nomad.3.8.1/src/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/nomad.3.8.1/ext/sgtelib/src/',
		'include/nomad.3.8.1/ext/sgtelib/src/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/lsape.5/include/', 'include/lsape.5/include/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/libsvm.3.22/', 'include/libsvm.3.22/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/fann.2.2.0/include/',
		'include/fann.2.2.0/include/'
	)
	# Copy-paste libs:
	print('Copying includes...')
	copy_paste_with_logs(
		'include/gedlib-master/ext/nomad.3.8.1/lib/', 'lib/nomad.3.8.1/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/libsvm.3.22/', 'lib/libsvm.3.22/'
	)
	copy_paste_with_logs(
		'include/gedlib-master/ext/fann.2.2.0/lib/', 'lib/fann.2.2.0/'
	)

	print('done!')
	print()


def check_and_include_glibc():
	def get_glibc_version():
		"""
		Authors
		-------
		Linlin Jia, ChatGPT 3.5 (2023.04.27)
		"""
		# Get the operating system name
		os_name = platform.system()

		# Command to get the version of glibc
		if os_name == 'Linux':
			# Linux (Ubuntu, CentOS)
			cmd = 'ldd --version | head -n 1'
		elif os_name == 'Darwin':
			# macOS
			cmd = 'otool -L /usr/lib/libc.dylib | head -n 1'
		elif os_name == 'Windows':
			# Windows
			cmd = 'dumpbin /dependents C:\\Windows\\System32\\msvcrt.dll | findstr /C:"msvcrt.dll"'
		else:
			raise SystemError('Unsupported OS: %s.' % os_name)

		# Run the command and capture the output
		try:
			output = subprocess.check_output(cmd, shell=True)
		except subprocess.CalledProcessError:
			print('glibc is not installed.')
			return None

		# Extract the version number from the output
		output_str = output.decode('utf-8')
		if os_name == 'Linux':
			# Linux (Ubuntu, CentOS)
			version_str = output_str.split()[-1]
		elif os_name == 'Darwin':
			# macOS
			version_str = output_str.split()[0].split('.')[0]
		elif os_name == 'Windows':
			# Windows
			version_str = output_str.split()[2]
		else:
			raise SystemError('Unsupported OS: %s.' % os_name)

		return version_str


	# Check the current version of glibc:
	glibc_version = get_glibc_version()
	print(f"glibc version: {glibc_version}")

	if glibc_version is None or float(glibc_version) < 2.27:
		# Include the glibc 2.27 coming along with GEDLIB:
		return True
	else:
		return False


def clean_previous_build():
	# clean previous build
	print()
	print('Cleaning previous build...')
	for name in os.listdir():
		if (name.startswith('gedlibpy') and not (
				name.endswith('.pyx') or name.endswith('.pxd'))):
			os.remove(name)
		if name == 'build':
			shutil.rmtree(name)
	print('Done!')
	print()


# def get_extensions(include_glibc):
def get_extensions():
	exts = [
		Extension(
			"gedlibpy",
			# sources=["gedlibpy.pyx", "src/GedLibBind.cpp"],
			sources=["gedlibpy.pyx"],
			include_dirs=[
				"src", "include", "include/lsape.5/include",
				"include/eigen.3.3.4/Eigen",
				"include/nomad.3.8.1/src",
				"include/nomad.3.8.1/ext/sgtelib/src",
				"include/libsvm.3.22",
				"include/fann.2.2.0/include", "include/boost.1.69.0"
			],
			library_dirs=[
				"lib/fann.2.2.0", "lib/libsvm.3.22",  # "lib/gedlib",
				"lib/nomad.3.8.1"
			],
			libraries=["doublefann", "sgtelib", "svm", "nomad"],
			language="c++",
			extra_compile_args=["-std=c++11"],
			extra_link_args=["-std=c++11"]
		)
	]
	return exts


def remove_includes():
	print()
	print('Deleting includes...')
	name = os.path.join(os.getcwd(), 'include/')
	shutil.rmtree(name)
	print('Done!')
	print()


if __name__ == '__main__':
	args = parse_args()
	if args.use_existing_gedlib == 'false':
		# Download GEDLIB and unpack it:
		get_gedlib()
	else:
		# Check if GEDLIB already exists:
		check_gedlib()
	if args.build_gedlib == 'true':
		# Install GEDLIB:
		install_gedlib()
	# Copy-Paste includes and libs of GEDLIB:
	copy_gedlib_includes_and_libs()
	# # Deal with GLIBC library:
	# include_glibc = check_and_include_glibc()

	# clean previous build:
	clean_previous_build()

	print()
	print('Start building...')
	# Build gedlibpy:
	# extensions = get_extensions(include_glibc)
	extensions = get_extensions()
	with open("README.rst", "r") as fh:
		long_description = fh.read()

	# Attention: setup function can not be put inside a function!
	setup(
		ext_modules=cythonize(
			extensions,
			compiler_directives={'language_level': '3'}
		),
		name="gedlibpy",
		author="Lambert Natacha and Linlin Jia",
		author_email="linlin.jia@unibe.ch",
		description="A Python wrapper library for C++ library GEDLIB of graph edit distances",
		long_description=long_description,
		long_description_content_type="text/markdown",
		project_urls={
			# 'Documentation': 'https://graphkit-learn.readthedocs.io',
			'Source': 'https://github.com/jajupmochi/graphkit-learn/tree/master/gklearn/gedlib',
			'Tracker': 'https://github.com/jajupmochi/graphkit-learn/issues',
		},
		url="https://github.com/jajupmochi/graphkit-learn/tree/master/gklearn/gedlib",
		classifiers=[
			"Programming Language :: Python :: 3",
			"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
			"Operating System :: OS Independent",
			'Intended Audience :: Science/Research',
			'Intended Audience :: Developers',
		],
		include_dirs=[numpy.get_include()]
	)

	# List generated files:
	print()
	print('The following files are generated:')
	for name in os.listdir():
		if (name.startswith('gedlibpy') and not (
				name.endswith('.pyx') or name.endswith('.pxd'))):
			print(name)

	print()
	print('Build completed!')
	print()

	if args.develop_mode == 'false':
		# Remove GEDLIB include files:
		remove_includes()

# Commande Bash : python setup.py build_ext --inplace



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  1 15:12:31 2020

@author: ljia
"""

def constant_node_costs(edit_cost_name):
	if edit_cost_name == 'NON_SYMBOLIC' or edit_cost_name == 'LETTER2' or edit_cost_name == 'LETTER':
		return False
	elif edit_cost_name == 'CONSTANT':
		return True
	else:
		raise Exception('Can not recognize the given edit cost. Possible edit costs include: "NON_SYMBOLIC", "LETTER", "LETTER2", "CONSTANT".')
#	 elif edit_cost_name != '':
# # 		throw ged::Error("Invalid dataset " + dataset + ". Usage: ./median_tests <AIDS|Mutagenicity|Letter-high|Letter-med|Letter-low|monoterpenoides|SYNTHETICnew|Fingerprint|COIL-DEL>");
#		 return False
	# return True
	
	
def mge_options_to_string(options):
	opt_str = ' '
	for key, val in options.items():
		if key == 'init_type':
			opt_str += '--init-type ' + str(val) + ' '
		elif key == 'random_inits':
			opt_str += '--random-inits ' + str(val) + ' '
		elif key == 'randomness':
			opt_str += '--randomness ' + str(val) + ' '
		elif key == 'verbose':
			opt_str += '--stdout ' + str(val) + ' '
		elif key == 'parallel':
			opt_str += '--parallel ' + ('TRUE' if val else 'FALSE') + ' '
		elif key == 'update_order':
			opt_str += '--update-order ' + ('TRUE' if val else 'FALSE') + ' '
		elif key == 'sort_graphs':
			opt_str += '--sort-graphs ' + ('TRUE' if val else 'FALSE') + ' '
		elif key == 'refine':
 			opt_str += '--refine ' + ('TRUE' if val else 'FALSE') + ' '
		elif key == 'time_limit':
 			opt_str += '--time-limit ' + str(val) + ' '
		elif key == 'max_itrs':
			opt_str += '--max-itrs ' + str(val) + ' '
		elif key == 'max_itrs_without_update':
			opt_str += '--max-itrs-without-update ' + str(val) + ' '
		elif key == 'seed':
			opt_str += '--seed ' + str(val) + ' '
		elif key == 'epsilon':
			opt_str += '--epsilon ' + str(val) + ' '
		elif key == 'inits_increase_order':
			opt_str += '--inits-increase-order ' + str(val) + ' '
		elif key == 'init_type_increase_order':
			opt_str += '--init-type-increase-order ' + str(val) + ' '
		elif key == 'max_itrs_increase_order':
			opt_str += '--max-itrs-increase-order ' + str(val) + ' '
# 		else:
# 			valid_options = '[--init-type <arg>] [--random_inits <arg>] [--randomness <arg>] [--seed <arg>] [--verbose <arg>] '
# 			valid_options += '[--time_limit <arg>] [--max_itrs <arg>] [--epsilon <arg>] '
# 			valid_options += '[--inits_increase_order <arg>] [--init_type_increase_order <arg>] [--max_itrs_increase_order <arg>]'
# 			raise Exception('Invalid option "' + key + '". Options available = "' + valid_options + '"')

	return opt_str


#export LD_LIBRARY_PATH=.:/export/home/lambertn/Documents/gedlibpy/lib/fann/:/export/home/lambertn/Documents/gedlibpy/lib/libsvm.3.22:/export/home/lambertn/Documents/gedlibpy/lib/nomad

#Pour que "import script" trouve les librairies qu'a besoin GedLib
#Equivalent à définir la variable d'environnement LD_LIBRARY_PATH sur un bash
import librariesImport
import gedlibpy
import networkx as nx

def init() :
    print("List of Edit Cost Options : ")
    for i in gedlibpy.list_of_edit_cost_options :
        print (i)
    print("")

    print("List of Method Options : ")
    for j in gedlibpy.list_of_method_options :
        print (j)
    print("")

    print("List of Init Options : ")
    for k in gedlibpy.list_of_init_options :
        print (k)
    print("")
    
init()

def afficheMatrix(mat) :
    for i in mat :
        line = ""
        for j in i :
            line+=str(j)
            line+=" "
        print(line)

def createNxGraph() :
    G = nx.Graph()
    G.add_node("1", chem = "C")
    G.add_node("2", chem = "0")
    G.add_edge("1", "2", valence = "1")
    G.add_node("3", chem = "N")
    G.add_node("4", chem = "C")
    G.add_edge("3", "4", valence = "1")
    G.add_edge("3", "2", valence = "1")
    return G

#G = createNxGraph()

def addGraphTest() :
    gedlibpy.restart_env()
    gedlibpy.load_GXL_graphs('include/gedlib-master/data/datasets/Mutagenicity/data/', 'collections/MUTA_10.xml')

    currentID = gedlibpy.add_graph()
    print(currentID)
    
    gedlibpy.add_node(currentID, "_1", {"chem" : "C"})
    gedlibpy.add_node(currentID, "_2", {"chem" : "O"})
    gedlibpy.add_edge(currentID,"_1", "_2",  {"valence": "1"} )

    listID = gedlibpy.get_all_graph_ids()
    print(listID)
    print(gedlibpy.get_graph_node_labels(10))
    print(gedlibpy.get_graph_edges(10))
    
    for i in listID : 
        print(gedlibpy.get_graph_node_labels(i))
        print(gedlibpy.get_graph_edges(i))

#addGraphTest()

def shortTest() :
    gedlibpy.restart_env()
    
    print("Here is the mini Python function !")
    
    gedlibpy.load_GXL_graphs("include/gedlib-master/data/datasets/Mutagenicity/data/", "include/gedlib-master/data/collections/Mutagenicity.xml")
    listID = gedlibpy.get_all_graph_ids()
    gedlibpy.set_edit_cost("CHEM_1")

    gedlibpy.init()

    gedlibpy.set_method("BIPARTITE", "")
    gedlibpy.init_method()

    g = listID[0]
    h = listID[1]

    gedlibpy.run_method(g,h)

    print("Node Map : ", gedlibpy.get_node_map(g,h))
    print("Assignment Matrix : ")
    afficheMatrix(gedlibpy.get_assignment_matrix(g,h))
    print ("Upper Bound = " + str(gedlibpy.get_upper_bound(g,h)) + ", Lower Bound = " + str(gedlibpy.get_lower_bound(g,h)) + ", Runtime = " + str(gedlibpy.get_runtime(g,h)))

#shortTest()

def classiqueTest() :
    gedlibpy.restart_env()
    
    gedlibpy.load_GXL_graphs('include/gedlib-master/data/datasets/Mutagenicity/data/', 'collections/MUTA_10.xml')
    listID = gedlibpy.get_all_graph_ids()
    
    afficheId = ""
    for i in listID :
        afficheId+=str(i) + " "
    print("Number of graphs = " + str(len(listID)) + ", list of Ids = " + afficheId)

    gedlibpy.set_edit_cost("CHEM_1")

    gedlibpy.init()

    gedlibpy.set_method("IPFP", "")
    gedlibpy.init_method()

    g = listID[0]
    h = listID[0]

    gedlibpy.run_method(g,h)
    liste = gedlibpy.get_all_map(g,h)
    print("Forward map : " , gedlibpy.get_forward_map(g,h), ", Backward map : ", gedlibpy.get_backward_map(g,h))
    print("Node Map : ", gedlibpy.get_node_map(g,h))
    print ("Upper Bound = " + str(gedlibpy.get_upper_bound(g,h)) + ", Lower Bound = " + str(gedlibpy.get_lower_bound(g,h)) + ", Runtime = " + str(gedlibpy.get_runtime(g,h)))

#classiqueTest()

def nxTest(dataset) :
    gedlibpy.restart_env()
    
    for graph in dataset :
        gedlibpy.add_nx_graph(graph, "")
        
    listID = gedlibpy.get_all_graph_ids()
    gedlibpy.set_edit_cost("CHEM_1")
    gedlibpy.init()
    gedlibpy.set_method("IPFP", "")
    gedlibpy.init_method()

    print(listID)
    g = listID[0]
    h = listID[1]

    gedlibpy.run_method(g,h)

    print("Node Map : ", gedlibpy.get_node_map(g,h))
    print ("Upper Bound = " + str(gedlibpy.get_upper_bound(g,h)) + ", Lower Bound = " + str(gedlibpy.get_lower_bound(g,h)) + ", Runtime = " + str(gedlibpy.get_runtime(g,h)))

#dataset = [createNxGraph(), createNxGraph()]
#nxTest(dataset)

def LSAPETest(matrixCost) :
    result = gedlibpy.hungarian_LSAPE(matrixCost)
    print("Rho = ", result[0], " Varrho = ", result[1], " u = ", result[2], " v = ", result[3])

#LSAPETest([[2,3,4], [5,1,9], [7,10,3]])

def LSAPTest(matrixCost) :
    result = gedlibpy.hungarian_LSAP(matrixCost)
    print("Rho = ", result[0], " Varrho = ", result[1], " u = ", result[2], " v = ", result[3])

#LSAPETest([[2,3,4], [5,1,9], [7,10,3]])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 16:55:17 2020

@author: ljia

@references: 

	[1] S Vichy N Vishwanathan, Nicol N Schraudolph, Risi Kondor, and Karsten M Borgwardt. Graph kernels. Journal of Machine Learning Research, 11(Apr):1201–1242, 2010.
"""

from gklearn.kernels import SylvesterEquation, ConjugateGradient, FixedPoint, SpectralDecomposition


class RandomWalk(SylvesterEquation, ConjugateGradient, FixedPoint, SpectralDecomposition):
	
	
	def __init__(self, **kwargs):
		self._compute_method = kwargs.get('compute_method', None)
		self._compute_method = self._compute_method.lower()
		
		if self._compute_method == 'sylvester':
			self._parent = SylvesterEquation
		elif self._compute_method == 'conjugate':
			self._parent = ConjugateGradient
		elif self._compute_method == 'fp':
			self._parent = FixedPoint
		elif self._compute_method == 'spectral':
			self._parent = SpectralDecomposition
		elif self._compute_method == 'kon':
			raise Exception('This computing method is not completed yet.')
		else:
			raise Exception('This computing method does not exist. The possible choices inlcude: "sylvester", "conjugate", "fp", "spectral".')

		self._parent.__init__(self, **kwargs)
		
		
	def _compute_gm_series(self):
		return self._parent._compute_gm_series(self)


	def _compute_gm_imap_unordered(self):
		return self._parent._compute_gm_imap_unordered(self)
	
		
	def _compute_kernel_list_series(self, g1, g_list):
		return self._parent._compute_kernel_list_series(self, g1, g_list)

	
	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		return self._parent._compute_kernel_list_imap_unordered(self, g1, g_list)
	
	
	def _compute_single_kernel_series(self, g1, g2):
		return self._parent._compute_single_kernel_series(self, g1, g2)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 11 11:03:01 2020

@author: ljia
"""
import numpy as np
from sklearn.model_selection import ShuffleSplit
from sklearn.neighbors import KNeighborsClassifier
from sklearn.metrics import accuracy_score
from gklearn.utils.utils import get_graph_kernel_by_name
# from gklearn.preimage.utils import get_same_item_indices

def sum_squares(a, b):
	"""
	Return the sum of squares of the difference between a and b, aka MSE
	"""
	return np.sum([(a[i] - b[i])**2 for i in range(len(a))])


def euclid_d(x, y):
	"""
	1D euclidean distance
	"""
	return np.sqrt((x-y)**2)


def man_d(x, y):
	"""
	1D manhattan distance
	"""
	return np.abs((x-y))


def knn_regression(D_app, D_test, y_app, y_test, n_neighbors, verbose=True, text=None):

	from sklearn.neighbors import KNeighborsRegressor
	knn = KNeighborsRegressor(n_neighbors=n_neighbors, metric='precomputed')
	knn.fit(D_app, y_app)
	y_pred = knn.predict(D_app)
	y_pred_test = knn.predict(D_test.T)
	perf_app = np.sqrt(sum_squares(y_pred, y_app)/len(y_app))
	perf_test = np.sqrt(sum_squares(y_pred_test, y_test)/len(y_test))

	if (verbose):
		print("Learning error with {} train examples : {}".format(text, perf_app))
		print("Test error with {} train examples : {}".format(text, perf_test))

	return perf_app, perf_test


def knn_classification(d_app, d_test, y_app, y_test, n_neighbors, verbose=True, text=None):
	knn = KNeighborsClassifier(n_neighbors=n_neighbors, metric='precomputed')
	knn.fit(d_app, y_app)
	y_pred = knn.predict(d_app)
	y_pred_test = knn.predict(d_test.T)
	perf_app = accuracy_score(y_app, y_pred)
	perf_test = accuracy_score(y_test, y_pred_test)

	if (verbose):
		print("Learning accuracy with {} costs : {}".format(text, perf_app))
		print("Test accuracy with {} costs : {}".format(text, perf_test))
		
	return perf_app, perf_test
	

def knn_cv(dataset, kernel_options, trainset=None, n_neighbors=1, n_splits=50, test_size=0.9, verbose=True):
	'''
	Perform a knn classification cross-validation on given dataset.
	'''
# 	Gn = dataset.graphs
	y_all = dataset.targets
	
	# compute kernel distances.
	dis_mat = _compute_kernel_distances(dataset, kernel_options, trainset=trainset)
		
	
	rs = ShuffleSplit(n_splits=n_splits, test_size=test_size, random_state=0)
# 	train_indices = [[] for _ in range(n_splits)] 
# 	test_indices = [[] for _ in range(n_splits)]
# 	idx_targets = get_same_item_indices(y_all)
# 	for key, item in idx_targets.items():
# 		i = 0
# 		for train_i, test_i in rs.split(item): # @todo: careful when parallel.
# 			train_indices[i] += [item[idx] for idx in train_i]
# 			test_indices[i] += [item[idx] for idx in test_i]
# 			i += 1
	
	accuracies = []
# 	for trial in range(len(train_indices)):
# 		train_index = train_indices[trial]
# 		test_index = test_indices[trial] 
	for train_index, test_index in rs.split(y_all):
# 		print(train_index, test_index)
# 		G_app = [Gn[i] for i in train_index]
# 		G_test = [Gn[i] for i in test_index]
		y_app = [y_all[i] for i in train_index]
		y_test = [y_all[i] for i in test_index]
		
		N = len(train_index)
		
		d_app = dis_mat.copy()
		d_app = d_app[train_index,:]
		d_app = d_app[:,train_index]
		
		d_test = np.zeros((N, len(test_index)))
		
		for i in range(N):
			for j in range(len(test_index)):
				d_test[i, j] = dis_mat[train_index[i], test_index[j]]
				
		accuracies.append(knn_classification(d_app, d_test, y_app, y_test, n_neighbors, verbose=verbose, text=''))
		
	results = {}
	results['ave_perf_train'] = np.mean([i[0] for i in accuracies], axis=0)
	results['std_perf_train'] = np.std([i[0] for i in accuracies], axis=0, ddof=1)
	results['ave_perf_test'] = np.mean([i[1] for i in accuracies], axis=0)
	results['std_perf_test'] = np.std([i[1] for i in accuracies], axis=0, ddof=1)

	return results
		
		
def _compute_kernel_distances(dataset, kernel_options, trainset=None):
	graph_kernel = get_graph_kernel_by_name(kernel_options['name'], 
				  node_labels=dataset.node_labels,
				  edge_labels=dataset.edge_labels, 
				  node_attrs=dataset.node_attrs,
				  edge_attrs=dataset.edge_attrs,
				  ds_infos=dataset.get_dataset_infos(keys=['directed']),
				  kernel_options=kernel_options)
	
	gram_matrix, run_time = graph_kernel.compute(dataset.graphs, **kernel_options)

	dis_mat, _, _, _ = graph_kernel.compute_distance_matrix()
	
	if trainset is not None:
		gram_matrix_unnorm = graph_kernel.gram_matrix_unnorm
		

	return dis_mat


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 16 17:26:40 2020

@author: ljia
"""
	
def test_median_graph_estimator():
	from gklearn.utils import load_dataset
	from gklearn.ged.median import MedianGraphEstimator, constant_node_costs
	from gklearn.gedlib import librariesImport, gedlibpy
	from gklearn.preimage.utils import get_same_item_indices
	import multiprocessing

	# estimator parameters.
	init_type = 'MEDOID'
	num_inits = 1
	threads = multiprocessing.cpu_count()
	time_limit = 60000
	
	# algorithm parameters.
	algo = 'IPFP'
	initial_solutions = 1
	algo_options_suffix = ' --initial-solutions ' + str(initial_solutions) + ' --ratio-runs-from-initial-solutions 1 --initialization-method NODE '

	edit_cost_name = 'LETTER2'
	edit_cost_constants = [0.02987291, 0.0178211, 0.01431966, 0.001, 0.001]
	ds_name = 'Letter_high'
	
	# Load dataset.
	# dataset = '../../datasets/COIL-DEL/COIL-DEL_A.txt'
	dataset = '../../../datasets/Letter-high/Letter-high_A.txt'
	Gn, y_all, label_names = load_dataset(dataset)
	y_idx = get_same_item_indices(y_all)
	for i, (y, values) in enumerate(y_idx.items()):
		Gn_i = [Gn[val] for val in values]
		break
	
	# Set up the environment.
	ged_env = gedlibpy.GEDEnv()
	# gedlibpy.restart_env()
	ged_env.set_edit_cost(edit_cost_name, edit_cost_constant=edit_cost_constants)
	for G in Gn_i:
		ged_env.add_nx_graph(G, '')
	graph_ids = ged_env.get_all_graph_ids()
	set_median_id = ged_env.add_graph('set_median')
	gen_median_id = ged_env.add_graph('gen_median')
	ged_env.init(init_option='EAGER_WITHOUT_SHUFFLED_COPIES')
	
	# Set up the estimator.
	mge = MedianGraphEstimator(ged_env, constant_node_costs(edit_cost_name))
	mge.set_refine_method(algo, '--threads ' + str(threads) + ' --initial-solutions ' + str(initial_solutions) + ' --ratio-runs-from-initial-solutions 1')
	
	mge_options = '--time-limit ' + str(time_limit) + ' --stdout 2 --init-type ' + init_type
	mge_options += ' --random-inits ' + str(num_inits) + ' --seed ' + '1'  + ' --update-order TRUE --refine FALSE --randomness PSEUDO  --parallel TRUE '# @todo: std::to_string(rng())
	
	# Select the GED algorithm.
	algo_options = '--threads ' + str(threads) + algo_options_suffix
	mge.set_options(mge_options)
	mge.set_label_names(node_labels=label_names['node_labels'],
					  edge_labels=label_names['edge_labels'], 
					  node_attrs=label_names['node_attrs'], 
					  edge_attrs=label_names['edge_attrs'])
	mge.set_init_method(algo, algo_options)
	mge.set_descent_method(algo, algo_options)
	
	# Run the estimator.
	mge.run(graph_ids, set_median_id, gen_median_id)
	
	# Get SODs.
	sod_sm = mge.get_sum_of_distances('initialized')
	sod_gm = mge.get_sum_of_distances('converged')
	print('sod_sm, sod_gm: ', sod_sm, sod_gm)
	
	# Get median graphs.
	set_median = ged_env.get_nx_graph(set_median_id)
	gen_median = ged_env.get_nx_graph(gen_median_id)
	
	return set_median, gen_median


def test_median_graph_estimator_symb():
	from gklearn.utils import load_dataset
	from gklearn.ged.median import MedianGraphEstimator, constant_node_costs
	from gklearn.gedlib import librariesImport, gedlibpy
	from gklearn.preimage.utils import get_same_item_indices
	import multiprocessing

	# estimator parameters.
	init_type = 'MEDOID'
	num_inits = 1
	threads = multiprocessing.cpu_count()
	time_limit = 60000
	
	# algorithm parameters.
	algo = 'IPFP'
	initial_solutions = 1
	algo_options_suffix = ' --initial-solutions ' + str(initial_solutions) + ' --ratio-runs-from-initial-solutions 1 --initialization-method NODE '

	edit_cost_name = 'CONSTANT'
	edit_cost_constants = [4, 4, 2, 1, 1, 1]
	ds_name = 'MUTAG'
	
	# Load dataset.
	dataset = '../../../datasets/MUTAG/MUTAG_A.txt'
	Gn, y_all, label_names = load_dataset(dataset)
	y_idx = get_same_item_indices(y_all)
	for i, (y, values) in enumerate(y_idx.items()):
		Gn_i = [Gn[val] for val in values]
		break
	Gn_i = Gn_i[0:10]
	
	# Set up the environment.
	ged_env = gedlibpy.GEDEnv()
	# gedlibpy.restart_env()
	ged_env.set_edit_cost(edit_cost_name, edit_cost_constant=edit_cost_constants)
	for G in Gn_i:
		ged_env.add_nx_graph(G, '')
	graph_ids = ged_env.get_all_graph_ids()
	set_median_id = ged_env.add_graph('set_median')
	gen_median_id = ged_env.add_graph('gen_median')
	ged_env.init(init_option='EAGER_WITHOUT_SHUFFLED_COPIES')
	
	# Set up the estimator.
	mge = MedianGraphEstimator(ged_env, constant_node_costs(edit_cost_name))
	mge.set_refine_method(algo, '--threads ' + str(threads) + ' --initial-solutions ' + str(initial_solutions) + ' --ratio-runs-from-initial-solutions 1')
	
	mge_options = '--time-limit ' + str(time_limit) + ' --stdout 2 --init-type ' + init_type
	mge_options += ' --random-inits ' + str(num_inits) + ' --seed ' + '1'  + ' --update-order TRUE --refine FALSE --randomness PSEUDO --parallel TRUE '# @todo: std::to_string(rng())
	
	# Select the GED algorithm.
	algo_options = '--threads ' + str(threads) + algo_options_suffix
	mge.set_options(mge_options)
	mge.set_label_names(node_labels=label_names['node_labels'],
					  edge_labels=label_names['edge_labels'], 
					  node_attrs=label_names['node_attrs'], 
					  edge_attrs=label_names['edge_attrs'])
	mge.set_init_method(algo, algo_options)
	mge.set_descent_method(algo, algo_options)
	
	# Run the estimator.
	mge.run(graph_ids, set_median_id, gen_median_id)
	
	# Get SODs.
	sod_sm = mge.get_sum_of_distances('initialized')
	sod_gm = mge.get_sum_of_distances('converged')
	print('sod_sm, sod_gm: ', sod_sm, sod_gm)
	
	# Get median graphs.
	set_median = ged_env.get_nx_graph(set_median_id)
	gen_median = ged_env.get_nx_graph(gen_median_id)
	
	return set_median, gen_median


if _name_ == '_main_':
 	# set_median, gen_median = test_median_graph_estimator()
 	set_median, gen_median = test_median_graph_estimator_symb()